__all__ = ('ApplicationCommand', 'ApplicationCommandOption', 'ApplicationCommandOptionChoice',
     'ApplicationCommandPermission', 'ApplicationCommandPermissionOverwrite', )

import warnings

from ..bases import DiscordEntity, maybe_snowflake
from ..core import APPLICATION_COMMANDS, ROLES
from ..preconverters import preconvert_preinstanced_type
from ..utils import is_valid_application_command_name, DATETIME_FORMAT_CODE
from ..user import User, UserBase, ClientUserBase, create_partial_user_from_id
from ..role import Role, create_partial_role_from_id

from .preinstanced import ApplicationCommandOptionType, ApplicationCommandPermissionOverwriteTargetType, \
    ApplicationCommandTargetType, APPLICATION_COMMAND_CONTEXT_TARGET_TYPES

APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER = ApplicationCommandPermissionOverwriteTargetType.user
APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE = ApplicationCommandPermissionOverwriteTargetType.role

# ApplicationCommand
APPLICATION_COMMAND_LIMIT_GLOBAL = 100
APPLICATION_COMMAND_LIMIT_GUILD = 100
APPLICATION_COMMAND_LENGTH_MAX = 4000

# ApplicationCommand & ApplicationCommandOption
APPLICATION_COMMAND_NAME_LENGTH_MIN = 1
APPLICATION_COMMAND_NAME_LENGTH_MAX = 32
APPLICATION_COMMAND_DESCRIPTION_LENGTH_MIN = 2
APPLICATION_COMMAND_DESCRIPTION_LENGTH_MAX = 100
APPLICATION_COMMAND_OPTIONS_MAX = 25
APPLICATION_COMMAND_CHOICES_MAX = 25

# ApplicationCommandOptionChoice
APPLICATION_COMMAND_CHOICE_NAME_LENGTH_MIN = 1
APPLICATION_COMMAND_CHOICE_NAME_LENGTH_MAX = 100
APPLICATION_COMMAND_CHOICE_VALUE_LENGTH_MIN = 0
APPLICATION_COMMAND_CHOICE_VALUE_LENGTH_MAX = 100

# ApplicationCommandPermissionOverwrite
APPLICATION_COMMAND_PERMISSION_OVERWRITE_MAX = 10

class ApplicationCommand(DiscordEntity, immortal=True):
    """
    Represents a Discord slash command.
    
    Attributes
    ----------
    id : `int`
        The application command's id.
    allow_by_default : `bool`
        Whether the command is enabled by default for everyone who has `use_application_commands` permission.
    application_id : `int`
        The application command's application's id.
    description : `str`
        The command's description. It's length can be in range [2:100].
    name : `str`
        The name of the command. It's length can be in range [1:32].
    options : `None` or `list` of ``ApplicationCommandOption``
        The parameters of the command. It's length can be in range [0:25]. If would be set as empty list, instead is
        set as `None`.
    target_type : ``ApplicationCommandTargetType``
        The application command target's type describing where it shows up.
    
    Notes
    -----
    ``ApplicationCommand`` instances are weakreferable.
    """
    __slots__ = ('allow_by_default', 'application_id', 'description', 'name', 'options', 'target_type',)
    
    def __new__(cls, name, description=None, *, allow_by_default=True, options=None, target_type=None):
        """
        Creates a new ``ApplicationCommand`` instance with the given parameters.
        
        Parameters
        ----------
        name : `str`
            The name of the command. It's length can be in range [1:32].
        
        description : `None` or `str`, Optional
            The command's description. It's length can be in range [2:100].
        
        allow_by_default : `bool`, Optional (Keyword only)
            Whether the command is enabled by default for everyone who has `use_application_commands` permission.
            
            Defaults to `True`.
        
        options : `None` or (`list` or `tuple`) of ``ApplicationCommandOption``, Optional (Keyword only)
            The parameters of the command. It's length can be in range [0:25].
        
        target_type : `int`, ``ApplicationCommandTargetType``, Optional (Keyword only)
            The application command's target type.
            
            Defaults to `ApplicationCommandTargetType.chat`.
        
        Raises
        ------
        TypeError
            If `target_type` is neither `int`, nor ``ApplicationCommandTargetType`` instance.
        ValueError
            `description` cannot be `None` for application commands with non-context target.
        AssertionError
            - If `name` was not given as `str` instance.
            - If `name` length is out of range [1:32].
            - If `name` contains unexpected character.
            - If `description` was not given as `None` nor `str` instance.
            - If `description` length is out of range [1:100].
            - If `options` was not given neither as `None` nor as (`list` or `tuple`) of ``ApplicationCommandOption``
                instances.
            - If `options`'s length is out of range [0:25].
            - If `allow_by_default` was not given as `bool` instance.
        """
        if __debug__:
            if not isinstance(name, str):
                raise AssertionError(f'`name` can be given as `str` instance, got {name.__class__.__name__}.')
            
            name_length = len(name)
            if name_length < APPLICATION_COMMAND_NAME_LENGTH_MIN or name_length > APPLICATION_COMMAND_NAME_LENGTH_MAX:
                raise AssertionError(f'`name` length can be in range '
                    f'[{APPLICATION_COMMAND_NAME_LENGTH_MIN}:{APPLICATION_COMMAND_NAME_LENGTH_MAX}], got '
                    f'{name_length!r}; {name!r}.')
            
            if not is_valid_application_command_name(name):
                raise AssertionError(f'`name` contains an unexpected character; Got {name!r}.')
            
            if (description is not None):
                if not isinstance(description, str):
                    raise AssertionError(f'`description` can be given as `None` or `str` instance, got '
                        f'{description.__class__.__name__}.')
                
                description_length = len(description)
                if description_length < APPLICATION_COMMAND_DESCRIPTION_LENGTH_MIN or \
                        description_length > APPLICATION_COMMAND_DESCRIPTION_LENGTH_MAX:
                    raise AssertionError(f'`description` length can be in range '
                        f'[{APPLICATION_COMMAND_DESCRIPTION_LENGTH_MIN}:{APPLICATION_COMMAND_DESCRIPTION_LENGTH_MAX}], '
                        f'got {description_length!r}; {description!r}.')
            
            if not isinstance(allow_by_default, bool):
                raise AssertionError(f'`allow_by_default` can be given as `bool` instance, got '
                    f'{allow_by_default.__class__.__name__}.')
        
        if options is None:
            options_processed = None
        else:
            if __debug__:
                if not isinstance(options, (tuple, list)):
                    raise AssertionError(f'`options` can be given as `None` or (`list` or `tuple`) of '
                        f'`{ApplicationCommandOption.__name__}`, got {options.__class__.__name__}.')
            
            # Copy it
            options_processed = list(options)
            if options_processed:
                if __debug__:
                    if len(options_processed) > APPLICATION_COMMAND_OPTIONS_MAX:
                        raise AssertionError(f'`options` length can be in range '
                            f'[0:{APPLICATION_COMMAND_OPTIONS_MAX}], got {len(options_processed)!r}; {options!r}')
                    
                    for index, option in enumerate(options_processed):
                        if not isinstance(option, ApplicationCommandOption):
                            raise AssertionError(f'`options` was given either as `list` or `tuple`, but it\'s element '
                                f'At index {index!r} is not {ApplicationCommandOption.__name__} instance, but '
                                f'{option.__class__.__name__}.')
            
            else:
                options_processed = None
        
        if target_type is None:
            target_type = ApplicationCommandTargetType.chat
        else:
            target_type = preconvert_preinstanced_type(target_type, 'target_type', ApplicationCommandTargetType)
        
        if (target_type not in APPLICATION_COMMAND_CONTEXT_TARGET_TYPES):
            if (description is None):
                raise ValueError(f'`description` cannot be `None` for application commands with non-context target.')
        else:
            # We do not really care about them, we can just lose them, no problem.
            description = None
            options_processed = None
        
        self = object.__new__(cls)
        self.id = 0
        self.application_id = 0
        self.name = name
        self.description = description
        self.allow_by_default = allow_by_default
        self.options = options_processed
        self.target_type = target_type
        return self
    
    
    def add_option(self, option):
        """
        Adds a new option to the application command.
        
        Parameters
        ----------
        option : ``ApplicationCommandOption``
            The option to add.
        
        Returns
        -------
        self : ``ApplicationCommand``
        
        Raises
        ------
        AssertionError
            - If the entity is not partial.
            - If `option` is not ``ApplicationCommandOption`` instance.
            - If the ``ApplicationCommand`` has already `25` options.
        """
        if __debug__:
            if self.id != 0:
                raise AssertionError(f'{self.__class__.__name__}.add_option` can be only called on partial '
                    f'`{self.__class__.__name__}`-s, but was called on {self!r}.')
        
        if __debug__:
            if not isinstance(option, ApplicationCommandOption):
                raise AssertionError(f'`option` can be given as {ApplicationCommandOption.__name__} instance, got '
                    f'{option.__class__.__name__}.')
        
        options = self.options
        if options is None:
            self.options = options = []
        else:
            if __debug__:
                if len(options) >= APPLICATION_COMMAND_OPTIONS_MAX:
                    raise AssertionError(f'`option` cannot be added if the {ApplicationCommandOption.__name__} has '
                        f'already `{APPLICATION_COMMAND_OPTIONS_MAX}` options.')
        
        options.append(option)
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new ``ApplicationCommand`` from requested data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received application command data.
        
        Returns
        -------
        self : ``ApplicationCommand``
            The created application command instance.
        """
        application_command_id = int(data['id'])
        try:
            self = APPLICATION_COMMANDS[application_command_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = application_command_id
            self.application_id = int(data['application_id'])
            APPLICATION_COMMANDS[application_command_id] = self
        
            # Discord might not include attributes in edit data, so we will set them first to avoid unset attributes.
            self.description = None
            self.name = ''
            self.options = None
            self.allow_by_default = True
            self.target_type = ApplicationCommandTargetType.none
        
        self._update_attributes(data)
        return self
    
    
    def _update_attributes(self, data):
        """
        Updates the application command with the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received application command data.
        """
        try:
            description = data['description']
        except KeyError:
            pass
        else:
            if (description is not None) and (not description):
                description = None
            self.description = description
        
        try:
            self.name = data['name']
        except KeyError:
            pass
        
        try:
            option_datas = data['options']
        except KeyError:
            pass
        else:
            if (option_datas is None) or (not option_datas):
                options = None
            else:
                options = [ApplicationCommandOption.from_data(option_data) for option_data in option_datas]
            self.options = options
        
        try:
            self.allow_by_default = data['default_permission']
        except KeyError:
            pass
        
        try:
            target_type = data['type']
        except KeyError:
            pass
        else:
            self.target_type = ApplicationCommandTargetType.get(target_type)
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the application command with the given data and returns the updated attributes in a dictionary with the
        attribute names as the keys and their old value as the values.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received application command data.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            The updated attributes.
            
            Every item in the returned dict is optional and can contain the following ones:
            
            +-----------------------+---------------------------------------------------+
            | Keys                  | Values                                            |
            +=======================+===================================================+
            | description           | `None` or `str`                                   |
            +-----------------------+---------------------------------------------------+
            | allow_by_default      | `bool`                                            |
            +-----------------------+---------------------------------------------------+
            | name                  | `str`                                             |
            +-----------------------+---------------------------------------------------+
            | options               | `None` or `list` of ``ApplicationCommandOption``  |
            +-----------------------+---------------------------------------------------+
            | target_type           | ``ApplicationCommandTargetType``                  |
            +-----------------------+---------------------------------------------------+
        """
        old_attributes = {}
        
        try:
            description = data['description']
        except KeyError:
            pass
        else:
            if (description is not None) and (not description):
                description = None
            if self.description != description:
                old_attributes['description'] = self.description
                self.description = description
        
        try:
            name = data['name']
        except KeyError:
            pass
        else:
            if self.name != name:
                old_attributes['name'] = self.name
                self.name = name
        
        try:
            option_datas = data['options']
        except KeyError:
            pass
        else:
            if (option_datas is None) or (not option_datas):
                options = None
            else:
                options = [ApplicationCommandOption.from_data(option_data) for option_data in option_datas]
            
            if self.options != options:
                old_attributes['options'] = self.options
                self.options = options
        
        try:
            allow_by_default = data['default_permission']
        except KeyError:
            pass
        else:
            if self.allow_by_default != self.allow_by_default:
                old_attributes['allow_by_default'] = allow_by_default
                self.allow_by_default = allow_by_default
        
        try:
            target_type = data['type']
        except KeyError:
            pass
        else:
            target_type = ApplicationCommandTargetType.get(target_type)
            if (self.target_type is not target_type):
                old_attributes['target_type'] = self.target_type
                self.target_type = target_type
        
        
        return old_attributes
    
    
    def to_data(self):
        """
        Converts the application command to a json serializable object.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        data = {
            'name': self.name,
        }
        
        description = self.description
        if (description is not None):
            data['description'] = description
        
        options = self.options
        if (options is None):
            option_datas = []
        else:
            option_datas = [option.to_data() for option in options]
        
        data['options'] = option_datas
        
        # Always add this to data, so if we update the command with it, will be always updated.
        data['default_permission'] = self.allow_by_default
        
        data['type'] = self.target_type.value
        
        return data
    
    def __repr__(self):
        """Returns the application command's representation."""
        repr_parts = [
            '<', self.__class__.__name__,
        ]
        
        id_ = self.id
        if id_ == 0:
            repr_parts.append(' (partial)')
        else:
            repr_parts.append(' id=')
            repr_parts.append(repr(id_))
            repr_parts.append(', application_id=')
            repr_parts.append(repr(self.application_id))
        
        repr_parts.append(', name=')
        repr_parts.append(repr(self.name))
        
        target_type = self.target_type
        if (target_type is not ApplicationCommandTargetType.none):
            repr_parts.append(', target_type=')
            repr_parts.append(target_type.name)
            repr_parts.append(' (')
            repr_parts.append(repr(target_type.value))
            repr_parts.append(')')
        
        description = self.description
        if (description is not None):
            repr_parts.append(', description=')
            repr_parts.append(repr(self.description))
        
        if not self.allow_by_default:
            repr_parts.append(', allow_by_default=False')
        
        options = self.options
        if (options is not None):
            repr_parts.append(', options=[')
            
            index = 0
            limit = len(options)
            
            while True:
                option = options[index]
                index += 1
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    @property
    def partial(self):
        """
        Returns whether the application command is partial.
        
        Returns
        -------
        partial : `bool`
        """
        if self.id == 0:
            return True
        
        return False
    
    
    def __hash__(self):
        """Returns the application's hash value."""
        id_ = self.id
        if id_:
            return id_
        
        raise TypeError(f'Cannot hash partial {self.__class__.__name__} object.')
    
    
    @classmethod
    def _from_edit_data(cls, data, interaction_id, application_id):
        """
        Creates an application command with the given parameters after an application command edition took place.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Application command data returned by it's ``.to_data`` method.
        interaction_id : `int`
            The unique identifier number of the newly created application command.
        application_id : `int`
            The new application identifier number of the newly created application command.
        
        Returns
        -------
        self : ``ApplicationCommand``
            The newly created or updated application command.
        """
        try:
            self = APPLICATION_COMMANDS[interaction_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = interaction_id
            self.application_id = application_id
            APPLICATION_COMMANDS[interaction_id] = self
            
            # Discord might not include attributes in edit data, so we will set them first to avoid unset attributes.
            self.description = None
            self.name = ''
            self.options = None
            self.target_type = ApplicationCommandTargetType.none
        
        self._update_attributes(data)
        
        return self
    
    
    def copy(self):
        """
        Copies the ``ApplicationCommand`` instance.
        
        Returns
        -------
        new : ``ApplicationCommand``
            A copied new partial application command.
        """
        new = object.__new__(type(self))
        new.id = 0
        new.application_id = 0
        new.name = self.name
        new.description = self.description
        new.allow_by_default = self.allow_by_default
        
        options = self.options
        if (options is not None):
            options = [option.copy() for option in options]
        new.options = options
        
        new.target_type = self.target_type
        
        return new
    
    def __eq__(self, other):
        """Returns whether the two application commands are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # If both entity is not partial, leave instantly by comparing id.
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            if self_id == other_id:
                return True
            
            return False
        
        if self.name != other.name:
            return False
        
        if self.description != other.description:
            return False
        
        if self.allow_by_default != other.allow_by_default:
            return False
        
        if self.options != other.options:
            return False
        
        if (self.target_type is not other.target_type):
            return False
        
        return True
    
    
    def __ne__(self, other):
        """Returns whether the two application commands are different."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            if self_id == other_id:
                return False
            
            return True
        
        if self.name != other.name:
            return True
        
        if self.description != other.description:
            return True
        
        if self.allow_by_default != other.allow_by_default:
            return True
        
        if self.options != other.options:
            return True
        
        if (self.target_type is not other.target_type):
            return True
        
        return False
    
    
    @property
    def mention(self):
        """
        Returns the application command's mention.
        
        Returns
        -------
        mention : `str`
        """
        return f'</{self.name}:{self.id}>'
    
    
    @property
    def display_name(self):
        """
        Returns the application command's display name.
        
        Returns
        -------
        display_name : `str`
        """
        return self.name.lower().replace('_', '-')
    
    
    def __format__(self, code):
        """
        Formats the application command in a format string.
        
        Parameters
        ----------
        code : `str`
            The option on based the result will be formatted.
        
        Returns
        -------
        application_command : `str`
        
        Raises
        ------
        ValueError
            Unknown format code.
        
        Examples
        --------
        ```py
        >>> from hata import ApplicationCommand
        >>> application_command = ApplicationCommand('cake-lover', 'Sends a random cake recipe OwO')
        >>> application_command
        <ApplicationCommand partial name='cake-lover', description='Sends a random cake recipe OwO'>
        >>> # no code stands for `application_command.name`.
        >>> f'{application_command}'
        'CakeLover'
        >>> # 'd' stands for display name.
        >>> f'{application_command:d}'
        'cake-lover'
        >>> # 'm' stands for mention.
        >>> f'{application_command:m}'
        '</cake-lover:0>'
        >>> # 'c' stands for created at.
        >>> f'{application_command:c}'
        '2021-01-03 20:17:36'
        ```
        """
        if not code:
            return self.name
        
        if code == 'm':
            return self.mention
        
        if code == 'd':
            return self.display_name
        
        if code == 'c':
            return self.created_at.__format__(DATETIME_FORMAT_CODE)
        
        raise ValueError(f'Unknown format code {code!r} for object of type {self.__class__.__name__!r}')
    
    
    def __len__(self):
        """Returns the application command's length."""
        length = len(self.name) + len(self.description)
        
        options = self.options
        if (options is not None):
            for option in options:
                length += len(option)
        
        return length
    
    
    def is_context_command(self):
        """
        Returns whether the application command is a context command.
        
        Returns
        -------
        is_context_command : `bool`
        """
        return (self.target_type in APPLICATION_COMMAND_CONTEXT_TARGET_TYPES)
    
    
    def is_slash_command(self):
        """
        Returns whether the application command is a slash command.
        
        Returns
        -------
        is_slash_command : `bool`
        """
        return (self.target_type is ApplicationCommandTargetType.chat)


class ApplicationCommandOption:
    """
    An option of an ``ApplicationCommand``.
    
    Attributes
    ----------
    autocomplete : `bool`
        Whether the option supports auto completion.
        
        Mutually exclusive with the ``.choices``. Only applicable for string type parameters.
    
    channel_types : `None` or `tuple` of `int`
        The accepted channel types by the option.
        
        Only applicable if ``.type`` is set to `ApplicationCommandOptionType.channel`.
    
    choices : `None` or `list` of ``ApplicationCommandOptionChoice``
        Choices for `str` and `int` types for the user to pick from.
        
        Mutually exclusive with the ``.autocomplete``.
    
    default : `bool`
        Whether the option is the default one. Only one option can be `default`.
    description : `str`
        The description of the application command option. It's length can be in range [1:100].
    max_value : `None`, `int` or `float`
        The maximal value permitted for this option.
        
        Only Applicable for integer as `int`, or `as float options as `float`.
    
    min_value : `None`, `int` or `float`
        The minimum value permitted for this option.
        
        Only Applicable for integer as `int`, or `as float options as `float`.
    
    name : `str`
        The name of the application command option. It's length can be in range [1:32].
    options : `None` or `list` of ``ApplicationCommandOption``
        If the command's type is sub-command group type, then this nested option will be the parameters of the
        sub-command. It's length can be in range [0:25]. If would be set as empty list, instead is set as `None`.
    required : `bool`
        Whether the parameter is required. Defaults to `False`.
    type : ``ApplicationCommandOptionType``
        The option's type.
    """
    __slots__ = ('autocomplete', 'channel_types', 'choices', 'default', 'description', 'max_value', 'min_value',
        'name', 'options', 'required', 'type')
    
    def __new__(cls, name, description, type_, *, autocomplete=False, channel_types=None, default=False,
            required=False, max_value=None, min_value=None, choices=None, options=None):
        """
        Creates a new ``ApplicationCommandOption`` instance with the given parameters.
        
        Parameters
        ----------
        name : `str`
            The name of the command. It's length can be in range [1:32].
        description : `str`
            The command's description. It's length can be in range [2:100].
        type_ : `int` or ``ApplicationCommandOptionType``
            The application command option's type.
        autocomplete : `bool`
            Whether the option supports auto completion.
            
            Mutually exclusive with the `choices` parameter. Only applicable for string type parameters.
        
        channel_types : `None` or `iterable` of `int`, Optional (Keyword only)
            The accepted channel types by the option.
            
            Only applicable if ``.type`` is set to `ApplicationCommandOptionType.channel`.
        
        default : `bool`, Optional (Keyword only)
            Whether the option is the default one. Defaults to `False`.
        required : `bool`, Optional (Keyword only)
            Whether the parameter is required. Defaults to `False`.
        max_value : `None`, `int` or `float`
            The maximal value permitted for this option.
            
            Only Applicable for integer as `int`, or `as float options as `float`.
        
        min_value : `None`, `int` or `float`
            The minimum value permitted for this option.
            
            Only Applicable for integer as `int`, or `as float options as `float`.
            
        choices : `None` or (`list` or `tuple`) of ``ApplicationCommandOptionChoice``, Optional (Keyword only)
            The choices of the command for string or integer types. It's length can be in range [0:25].
            
            Mutually exclusive with the `autocomplete` parameter.
            
        options : `None` or (`list` or `tuple`) of ``ApplicationCommandOption``, Optional (Keyword only)
            The parameters of the command. It's length can be in range [0:25]. Only applicable for sub command groups.
        
        Raises
        ------
        TypeError
            - If `type_` was not given neither as `int` nor ``ApplicationCommandOptionType`` instance.
            - If `choices` was given meanwhile `type_` is neither string nor integer option type.
            - If `options` was given meanwhile `type_` is not a sub-command group option type.
            - If a choice's value's type not matched the expected type described `type_`.
            - If `channel_types` is neither `None` nor `iterable` of `int`.
            - If `max_value` is not the expected type defined by `type_`'s value.
            - If `min_value` is not the expected type defined by `type_`'s value.
        ValueError
            - If `type_` was given as `int` instance, but it do not matches any of the precreated
                ``ApplicationCommandOptionType``-s.
            - If `channel_types` contains an unknown channel type value.
            - If `max_value` is given, but it is not applicable for the given `type_`.
            - If `min_value` is given, but it is not applicable for the given `type_`.
        AssertionError
            - If `name` was not given as `str` instance.
            - If `name` length is out of range [1:32].
            - If `description` was not given as `str` instance.
            - If `description` length is out of range [1:100].
            - If `options` was not given neither as `None` nor as (`list` or `tuple`) of ``ApplicationCommandOption``
                instances.
            - If `options`'s length is out of range [0:25].
            - If `default` was not given as `bool` instance.
            - If `required` was not given as `bool` instance.
            - If `choices` was not given neither as `None` nor as (`list` or `tuple`) of
                ``ApplicationCommandOptionChoice`` instances.
            - If `choices`'s length is out of range [0:25].
            - If an option is a sub command group option.
            - If `channel_types` is given, but `type_` is not `ApplicationCommandOptionType.channel`.
            - If `autocomplete` is not `bool` instance.
            - If both `autocomplete` and `choices` are defined.
            - If `autocomplete` is defined, but the parameters's type is not string.
        """
        if __debug__:
            if not isinstance(name, str):
                raise AssertionError(f'`name` can be given as `str` instance, got {name.__class__.__name__}.')
            
            name_length = len(name)
            if name_length < APPLICATION_COMMAND_NAME_LENGTH_MIN or \
                    name_length > APPLICATION_COMMAND_NAME_LENGTH_MAX:
                raise AssertionError(f'`name` length can be in range '
                    f'[{APPLICATION_COMMAND_NAME_LENGTH_MIN}:{APPLICATION_COMMAND_NAME_LENGTH_MAX}], got '
                    f'{name_length!r}; {name!r}.')
        
            if not isinstance(description, str):
                raise AssertionError(f'`description` can be given as `str` instance, got '
                    f'{description.__class__.__name__}.')
            
            description_length = len(description)
            if description_length < APPLICATION_COMMAND_DESCRIPTION_LENGTH_MIN or \
                    description_length > APPLICATION_COMMAND_DESCRIPTION_LENGTH_MAX:
                raise AssertionError(f'`description` length can be in range '
                    f'[{APPLICATION_COMMAND_DESCRIPTION_LENGTH_MIN}:'
                    f'{APPLICATION_COMMAND_DESCRIPTION_LENGTH_MAX}], got {description_length!r}; '
                    f'{description!r}.')
        
        type_ = preconvert_preinstanced_type(type_, 'type_', ApplicationCommandOptionType)
        
        if __debug__:
            if not isinstance(default, bool):
                raise AssertionError(f'`default` can be given as `bool` instance, got {default.__class__.__name__}.')
            
            if not isinstance(required, bool):
                raise AssertionError(f'`required` can be given as `bool` instance, got {required.__class__.__name__}.')
        
        if choices is None:
            choices_processed = None
        else:
            if __debug__:
                if not isinstance(choices, (tuple, list)):
                    raise AssertionError(f'`choices` can be given as `None` or (`list` or `tuple`) of '
                        f'`{ApplicationCommandOptionChoice.__name__}`, got {choices.__class__.__name__}.')
            
            choices_processed = list(choices)
            
            if __debug__:
                if len(choices_processed) > APPLICATION_COMMAND_CHOICES_MAX:
                    raise AssertionError(f'`choices` length can be in range '
                        f'[0:{APPLICATION_COMMAND_CHOICES_MAX}], got {len(choices_processed)!r}; '
                        f'{choices!r}')
                
                for index, choice in enumerate(choices_processed):
                    if not isinstance(choice, ApplicationCommandOptionChoice):
                        raise AssertionError(f'`choices` was given either as `list` or `tuple`, but it\'s element '
                            f'At index {index!r} is not {ApplicationCommandOptionChoice.__name__} instance, but '
                            f'{choice.__class__.__name__}; got {choices!r}.')
            
            if not choices_processed:
                choices_processed = None
        
        if options is None:
            options_processed = None
        else:
            if __debug__:
                if not isinstance(options, (tuple, list)):
                    raise AssertionError(f'`options` can be given as `None` or (`list` or `tuple`) of '
                        f'`{ApplicationCommandOption.__name__}`, got {options.__class__.__name__}.')
            
            # Copy it
            options_processed = list(options)
            
            if __debug__:
                if len(options_processed) > APPLICATION_COMMAND_OPTIONS_MAX:
                    raise AssertionError(f'`options` length can be in range '
                        f'[0:{APPLICATION_COMMAND_OPTIONS_MAX}], got {len(options_processed)!r}; {options!r}')
                
                for index, option in enumerate(options_processed):
                    if not isinstance(option, ApplicationCommandOption):
                        raise AssertionError(f'`options` was given either as `list` or `tuple`, but it\'s element '
                            f'At index {index!r} is not {ApplicationCommandOption.__name__} instance, but '
                            f'{option.__class__.__name__}; got {options!r}.')
                    
                    if option.type is ApplicationCommandOptionType.sub_command_group:
                        raise AssertionError(f'`options` element {index}\'s type is cub-command group option, but'
                             f'sub-command groups cannot be added under sub-command groups; got {options!r}.')
            
            if not options_processed:
                options_processed = None
        
        if (choices_processed is not None):
            if type_ is ApplicationCommandOptionType.string:
                expected_choice_type = str
            elif type_ is ApplicationCommandOptionType.integer:
                expected_choice_type = int
            elif type_ is ApplicationCommandOptionType.float:
                expected_choice_type = float
            else:
                raise TypeError(f'`choices` is bound to string, integer and float option type, got '
                    f'choices={choices!r}, type={type_!r}.')
            
            for index, choice in enumerate(choices):
                if not isinstance(choice.value, expected_choice_type):
                    raise TypeError(f'`choices` element\'s {index!r} value\'s type is not '
                        f'`{expected_choice_type.__name__}` as expected from the received command option type: '
                        f'{type_!r}')
                pass
        
        if (channel_types is None):
            channel_types_processed = None
        else:
            channel_types_processed = None
            
            iterator = getattr(type(channel_types), '__iter__', None)
            if (iterator is None):
                raise TypeError(f'`channel_types` is neither `None` nor `iterable`, got '
                    f'{channel_types.__class__.__anme__}.')
            
            for channel_type in iterator(channel_types):
                if type(channel_type) is int:
                    pass
                elif isinstance(channel_type, int):
                    channel_type = int(channel_type)
                else:
                    raise TypeError(f'`channel_types` may include only `int` instances, got '
                        f'{channel_type.__class__.__name__}; {channel_type!r}.')
                
                if channel_types_processed is None:
                    channel_types_processed = set()
                
                channel_types_processed.add(channel_type)
        
            if channel_types_processed:
                channel_types_processed = tuple(sorted(channel_types_processed))
            else:
                channel_types_processed = None
        
        
        if (max_value is not None):
            if type_ is ApplicationCommandOptionType.integer:
                if not isinstance(max_value, int):
                    raise TypeError(f'`max_value` can be `int` type, if `type_` is defined as {type_!r}, got '
                        f'{max_value.__class__.__name__}; {max_value!r}.')
            
            elif type_ is ApplicationCommandOptionType.float:
                if not isinstance(max_value, float):
                    raise TypeError(f'`max_value` can be `float` type, if `type_` is defined as {type_!r}, got '
                        f'{max_value.__class__.__name__}; {max_value!r}.')
            
            else:
                raise ValueError(f'`max_value` is only meaningful if `type` is either '
                    f'{ApplicationCommandOptionType.integer!r}, or {ApplicationCommandOptionType.float!r}, got '
                    f'type_={type!r}; max_value={max_value!r}.'
                )
        
        if (min_value is not None):
            if type_ is ApplicationCommandOptionType.integer:
                if not isinstance(min_value, int):
                    raise TypeError(f'`min_value` can be `int` type, if `type_` is defined as {type_!r}, got '
                        f'{min_value.__class__.__name__}; {min_value!r}.')
            
            elif type_ is ApplicationCommandOptionType.float:
                if not isinstance(min_value, float):
                    raise TypeError(f'`min_value` can be `float` type, if `type_` is defined as {type_!r}, got '
                        f'{min_value.__class__.__name__}; {min_value!r}.')
            
            else:
                raise ValueError(f'`min_value` is only meaningful if `type` is either '
                    f'{ApplicationCommandOptionType.integer!r}, or {ApplicationCommandOptionType.float!r}, got '
                    f'type_={type!r}; min_value={min_value!r}.'
                )
        
        if __debug__:
            if (channel_types_processed is not None) and (type_ is not ApplicationCommandOptionType.channel):
                raise AssertionError(f'`channel_types` is only meaningful if `type_` is '
                    f'`{ApplicationCommandOptionType.__name__}.channel`.')
        
            if not isinstance(autocomplete, bool):
                raise AssertionError(f'`autocomplete` can be `bool` instance, got {autocomplete.__class__.__name__}.')
            
            if autocomplete:
                if (choices_processed is not None):
                    raise AssertionError(f'`autocomplete` and `choices` parameters are mutually exclusive.')
                
                if (type_ is not ApplicationCommandOptionType.string):
                    raise AssertionError(f'`autocomplete` is only available for string option type, got {type_!r}')
        
        self = object.__new__(cls)
        self.name = name
        self.description = description
        self.type = type_
        self.default = default
        self.required = required
        self.choices = choices_processed
        self.options = options_processed
        self.channel_types = channel_types_processed
        self.autocomplete = autocomplete
        self.max_value = max_value
        self.min_value = min_value
        return self
    
    
    def add_option(self, option):
        """
        Adds a new option to the application command option.
        
        Parameters
        ----------
        option : ``ApplicationCommandOption``
            The option to add.
        
        Returns
        -------
        self : ``ApplicationCommandOption``
        
        Raises
        ------
        TypeError
            If the source application command's type is not a sub-command group type.
        AssertionError
            - If `option` is not ``ApplicationCommandOption`` instance.
            - If the ``ApplicationCommandOption`` has already `25` options.
            - If `option` is a sub command group option.
        """
        if self.type is not ApplicationCommandOptionType.sub_command_group:
            raise TypeError(f'`option` can be added only if the command option\s type is sub command option, '
                f'got option={option!r}, self={self!r}.')
        
        if __debug__:
            if not isinstance(option, ApplicationCommandOption):
                raise AssertionError(f'`option` can be given as {ApplicationCommandOption.__name__} instance, got '
                    f'{option.__class__.__name__}.')
        
            if option.type is ApplicationCommandOptionType.sub_command_group:
                raise AssertionError(f'`option`\'s type is sub-command group option, but sub-command groups cannot be '
                    f'added under sub-command groups; got {option!r}.')
        
        options = self.options
        if options is None:
            self.options = options = []
        else:
            if __debug__:
                if len(options) >= APPLICATION_COMMAND_OPTIONS_MAX:
                    raise AssertionError(f'`option` cannot be added if the {ApplicationCommandOption.__name__} has '
                        f'already `{APPLICATION_COMMAND_OPTIONS_MAX}` options.')
        
        options.append(option)
        return self
    
    
    def add_choice(self, choice):
        """
        Adds a ``ApplicationCommandOptionChoice`` to the application command option.
        
        Parameters
        ----------
        choice : ``ApplicationCommandOptionChoice`` or `tuple` (`str`, (`str`, `int`, `float`))
            The choice to add.
        
        Returns
        -------
        self : ``ApplicationCommandOption``
        
        Raises
        ------
        TypeError
            - If the source application command's type is not a string, int nor float group type.
            - If the `choice`'s value's type is not the expected one by the command option's type.
            - If `choice`'s type is neither ``ApplicationCommandOptionChoice`` nor a `tuple` representing it's `.name`
                nad `.value`.
        AssertionError
            If the application command option has already `25` choices.
        """
        if isinstance(choice, ApplicationCommandOptionChoice):
            pass
        elif isinstance(choice, tuple):
            if len(choice) != 2:
                raise TypeError(f'If `choice` is given as `tuple` it\'s length should be `2` representing a '
                    f'{ApplicationCommandOptionChoice.__name__}\s `.name` and `.value`.')
            
            choice = ApplicationCommandOptionChoice(*choice)
        else:
            raise TypeError(f'`choice` can be given as {ApplicationCommandOptionChoice.__name__} instance or a `tuple` '
                f'representing one with i\'s respective `.name` and `.value` as it\'s elements, got '
                f'{choice.__class__.__name__}.')
        
        type_ = self.type
        if type_ is ApplicationCommandOptionType.string:
            expected_choice_type = str
        elif type_ is ApplicationCommandOptionType.integer:
            expected_choice_type = int
        elif type_ is ApplicationCommandOptionType.float:
            expected_choice_type = float
        else:
            raise TypeError(f'`choice` is bound to string, integer and float choice type, got choice={choice!r}, '
                f'self={self!r}.')
        
        if not isinstance(choice.value, expected_choice_type):
            raise TypeError(f'`choice` value\'s type is not `{expected_choice_type.__name__}` as expected from the '
                f'received command choice type: {type_!r}')
        
        choices = self.choices
        if choices is None:
            self.choices = choices = []
        else:
            if __debug__:
                if len(choices) >= APPLICATION_COMMAND_CHOICES_MAX:
                    raise AssertionError(f'`choice` cannot be added if the {ApplicationCommandOption.__name__} has '
                        f'already `{APPLICATION_COMMAND_CHOICES_MAX}` choices.')
        
        choices.append(choice)
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new ``ApplicationCommandOption`` instance from the received data from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received application command option data.
        
        Returns
        -------
        self : ``ApplicationCommandOption``
            The created application command option.
        """
        self = object.__new__(cls)
        choice_datas = data.get('choices', None)
        if (choice_datas is None) or (not choice_datas):
            choices = None
        else:
            choices = [ApplicationCommandOptionChoice.from_data(choice_data) for choice_data in choice_datas]
        self.choices = choices
        
        self.default = data.get('default', False)
        
        self.description = data['description']
        self.name = data['name']
        
        option_datas = data.get('options', None)
        if (option_datas is None) or (not option_datas):
            options = None
        else:
            options = [ApplicationCommandOption.from_data(option_data) for option_data in option_datas]
        self.options = options
        
        self.required = data.get('required', False)
        
        channel_types = data.get('channel_types', None)
        if (channel_types is None) or (not channel_types):
            channel_types = None
        else:
            channel_types = tuple(sorted(channel_types))
        self.channel_types = channel_types
        
        self.autocomplete = data.get('autocomplete', False)
        
        self.min_value = data.get('min_value', None)
        
        self.max_value = data.get('max_value', None)
        
        self.type = ApplicationCommandOptionType.get(data['type'])
        return self
    
    
    def to_data(self):
        """
        Converts the application command option to a json serializable object.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        data = {
            'description' : self.description,
            'name' : self.name,
            'type' : self.type.value,
        }
        
        choices = self.choices
        if (choices is not None):
            data['choices'] = [choice.to_data() for choice in choices]
        
        
        if self.default:
            data['default'] = True
        
        options = self.options
        if (options is not None):
            data['options'] = [option.to_data() for option in options]
        
        if self.required:
            data['required'] = True
        
        channel_types = self.channel_types
        if (channel_types is not None):
            data['channel_types'] = channel_types
        
        if self.autocomplete:
            data['autocomplete'] = True
        
        min_value = self.min_value
        if (min_value is not None):
            data['min_value'] = min_value
        
        max_value = self.max_value
        if (max_value is not None):
            data['max_value'] = max_value
        
        return data
    
    
    def __repr__(self):
        """Returns the application command option's representation."""
        repr_parts = [
            '<', self.__class__.__name__,
            ', name=', repr(self.name),
        ]
        
        if self.autocomplete:
            repr_parts.append(' (auto completed)')
        
        repr_parts.append(', description=')
        repr_parts.append(repr(self.description))
        repr_parts.append(', type=')
        
        type_ = self.type
        repr_parts.append(repr(type_.value))
        repr_parts.append(' (')
        repr_parts.append(type_.name)
        repr_parts.append(')')
        
        min_value = self.min_value
        if (min_value is not None):
            repr_parts.append(', min_value=')
            repr_parts.append(repr(min_value))
        
        max_value = self.max_value
        if (max_value is not None):
            repr_parts.append(', max_value=')
            repr_parts.append(repr(max_value))
        
        if self.default:
            repr_parts.append(', default=True')
        
        if self.required:
            repr_parts.append(', required=True')
        
        choices = self.choices
        if (choices is not None):
            repr_parts.append(', choices=[')
            
            index = 0
            limit = len(choices)
            
            while True:
                choice = choices[index]
                index += 1
                repr_parts.append(repr(choice))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
        
        options = self.options
        if (options is not None):
            repr_parts.append(', options=[')
            
            index = 0
            limit = len(options)
            
            while True:
                option = options[index]
                index += 1
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        channel_types = self.channel_types
        if (channel_types is not None):
            repr_parts.append(', channel_types=[')
            
            index = 0
            limit = len(channel_types)
            
            while True:
                channel_type = channel_types[index]
                index += 1
                repr_parts.append(repr(channel_type))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def copy(self):
        """
        Copies the ``ApplicationCommandOption``.
        
        Returns
        -------
        new : ``ApplicationCommandOption``
        """
        new = object.__new__(type(self))
        
        choices = self.choices
        if (choices is not None):
            choices = choices.copy()
        new.choices = choices
        
        new.default = self.default
        new.description = self.description
        new.name = self.name
        
        options = self.options
        if (options is not None):
            options = [option.copy() for option in options]
        new.options = options
        
        new.required = self.required
        new.type = self.type
        
        channel_types = self.channel_types
        if (channel_types is not None):
            channel_types = tuple(channel_types)
        new.channel_types = channel_types
        
        new.autocomplete = self.autocomplete
        
        new.min_value = self.min_value
        new.max_value = self.max_value
        
        return new
    
    def __eq__(self, other):
        """Returns whether the two options are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.choices != other.choices:
            return False
        
        if self.default != other.default:
            return False
        
        if self.description != other.description:
            return False
        
        if self.name != other.name:
            return False
        
        if self.options != other.options:
            return False
        
        if self.required != other.required:
            return False
        
        if self.type is not other.type:
            return False
        
        if self.channel_types != other.channel_types:
            return False
        
        if self.autocomplete != other.autocomplete:
            return False
        
        if self.min_value != other.min_value:
            return False
        
        if self.max_value != other.max_value:
            return False
        
        return True
    
    def __len__(self):
        """Returns the application command option's length."""
        length = len(self.name) + len(self.description)
        
        choices = self.choices
        if (choices is not None):
            for choice in choices:
                length += len(choice)
        
        options = self.options
        if (options is not None):
            for option in options:
                length += len(option)
        
        return length


class ApplicationCommandOptionChoice:
    """
    A choice of a ``ApplicationCommandOption``.
    
    Attributes
    ----------
    name : `str`
        The choice's name. It's length can be in range [1:100].
    value : `str`, `int` or `float`
        The choice's value.
    """
    __slots__ = ('name', 'value')
    
    def __new__(cls, name, value):
        """
        Creates a new ``ApplicationCommandOptionChoice`` instance with the given parameters.
        
        Parameters
        ----------
        name : `str`
            The choice's name. It's length can be in range [1:100].
        value : `str`, `int` or `float`
            The choice's value.
        
        Raises
        ------
        AssertionError
            - If `name` is not `str` instance.
            - If `name`'s length is out of range [1:100].
            - If `value` is neither `str`, `int` nor `float` instance.
            - If `value` is `str` and it's length is out of range [0:100].
        """
        if __debug__:
            if not isinstance(name, str):
                raise AssertionError(f'`name` can be given as `str` instance, got {name.__class__.__name__}.')
            
            name_length = len(name)
            if name_length < APPLICATION_COMMAND_CHOICE_NAME_LENGTH_MIN or \
                    name_length > APPLICATION_COMMAND_CHOICE_NAME_LENGTH_MAX:
                raise AssertionError(f'`name` length can be in range '
                    f'[{APPLICATION_COMMAND_CHOICE_NAME_LENGTH_MIN}:{APPLICATION_COMMAND_CHOICE_NAME_LENGTH_MAX}], '
                    f'got {name_length!r}; {name!r}.')
            
            if isinstance(value, int):
                pass
            elif isinstance(value, str):
                value_length = len(value)
                if value_length < APPLICATION_COMMAND_CHOICE_VALUE_LENGTH_MIN or \
                        value_length > APPLICATION_COMMAND_CHOICE_VALUE_LENGTH_MAX:
                    raise AssertionError(f'`value` length` can be in range '
                        f'[{APPLICATION_COMMAND_CHOICE_VALUE_LENGTH_MIN}:{APPLICATION_COMMAND_CHOICE_NAME_LENGTH_MAX}]'
                        f'got {value_length!r}; {value!r}.')
            elif isinstance(value, float):
                pass
            else:
                raise AssertionError(f'`value` type can be either `str`, `int` or `float`, '
                    f'got {value.__class__.__name__}.')
        
        self = object.__new__(cls)
        self.name = name
        self.value = value
        return self
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new ``ApplicationCommandOptionChoice`` instance from the received data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            The received application command option choice data.
        
        Returns
        -------
        self : ``ApplicationCommandOptionChoice``
            The created choice.
        """
        self = object.__new__(cls)
        self.name = data['name']
        self.value = data['value']
        return self
    
    def to_data(self):
        """
        Converts the application command option choice to a json serializable object.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        return {
            'name': self.name,
            'value': self.value,
        }
    
    def __repr__(self):
        """Returns the application command option choice's representation."""
        return f'<{self.__class__.__name__} name={self.name!r}, value={self.value!r}>'
    
    def __eq__(self, other):
        """Returns whether the two choices are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.name != other.name:
            return False
        
        if self.value != other.value:
            return False
        
        return True
    
    def __len__(self):
        """Returns the application command choice's length."""
        length = len(self.name)
        value = self.value
        if isinstance(value, str):
            length += len(value)
        
        return length


class ApplicationCommandPermission:
    """
    Stores am ``ApplicationCommand``'s overwrites.
    
    Attributes
    ----------
    application_command_id : `int`
        The identifier of the respective ``ApplicationCommand``.
    application_id : `int`
        The application command's application's identifier.
    guild_id : `int`
        The identifier of the respective guild.
    permission_overwrites : `None` or `list` of ``ApplicationCommandPermissionOverwrite``
        The application command permissions overwrites relating to the respective application command in the guild.
    """
    __slots__ = ('application_command_id', 'application_id', 'guild_id', 'permission_overwrites')
    
    def __new__(cls, application_command, *, permission_overwrites=None, overwrites=None):
        """
        Creates a new ``ApplicationCommandPermission`` instance from the given parameters.
        
        Parameters
        ----------
        application_command : ``ApplicationCommand`` or `int`
            The application command's identifier.
        permission_overwrites : `None` or (`list`, `set`, `tuple`) of ``ApplicationCommandPermissionOverwrite`
                , Optional (Keyword only)
            Overwrites for the application command.
        
        Raises
        ------
        TypeError
            - If `application_command` was not given neither as ``ApplicationCommand`` nor as `int` instance.
        AssertionError
            - If `permission_overwrites` was not give neither as `None`, `list`, `set` or `tuple`.
            - If `permission_overwrites` contains a non ``ApplicationCommandPermissionOverwrite`` element.
            - If `permission_overwrites` length is over `10`.
        """
        if (overwrites is not None):
            warnings.warn(
                f'`ApplicationCommandPermission.__new__`\'s `overwrites` parameter is deprecated, '
                f'and will be removed in 2021 November. '
                f'Please use `permission_overwrites` instead.',
                FutureWarning)
            
            permission_overwrites = overwrites
        
        if isinstance(application_command, ApplicationCommand):
            application_command_id = application_command.id
        else:
            application_command_id = maybe_snowflake(application_command)
            if application_command_id is None:
                raise TypeError(f'`application_command` can be given as `{ApplicationCommand.__name__}`, or as `int` '
                    f'instance, got {application_command.__class__.__name__}.')
        
        if permission_overwrites is None:
            permission_overwrites_processed = None
        else:
            if __debug__:
                if not isinstance(permission_overwrites, (list, set, tuple)):
                    raise AssertionError(f'`permission_overwrites` can be given either as `None` or as `list`, '
                        f'`set`, `tuple`instance, got {permission_overwrites.__class__.__name__}.')
            
            permission_overwrites_processed = []
            
            for permission_overwrite in permission_overwrites:
                if __debug__:
                    if not isinstance(permission_overwrite, ApplicationCommandPermissionOverwrite):
                        raise AssertionError(f'`permission_overwrites` contains a non '
                            f'{ApplicationCommandPermissionOverwrite.__name__} element, got '
                            f'{overwrite.__class__.__name__}.')
                
                permission_overwrites_processed.append(permission_overwrite)
                
            
            if permission_overwrites_processed:
                if __debug__:
                    if len(overwrites) >= APPLICATION_COMMAND_PERMISSION_OVERWRITE_MAX:
                        raise AssertionError(f'`permission_overwrites` can contain up to '
                            f'`{APPLICATION_COMMAND_PERMISSION_OVERWRITE_MAX}` permission_overwrites, which is passed, '
                            f'got {len(permission_overwrites)!r}.')
            else:
                permission_overwrites_processed = None
        
        self = object.__new__(cls)
        self.application_command_id = application_command_id
        self.application_id = 0
        self.guild_id = 0
        self.permission_overwrites = permission_overwrites_processed
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new ``ApplicationCommandPermission`` instance.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Application command data.
        
        Returns
        -------
        self : ``ApplicationCommandPermission``
        """
        permission_overwrite_datas = data['permissions']
        if permission_overwrite_datas:
            permission_overwrites = [
                ApplicationCommandPermissionOverwrite.from_data(permission_overwrite_data) for
                permission_overwrite_data in permission_overwrite_datas
            ]
            
        else:
            permission_overwrites = None
        
        self = object.__new__(cls)
        self.application_command_id = int(data['id'])
        self.application_id = int(data['application_id'])
        self.guild_id = int(data['guild_id'])
        self.permission_overwrites = permission_overwrites
        return self
    
    
    def to_data(self):
        """
        Converts the application command permission to a json serializable object.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        data = {
            'id': self.application_command_id,
            'application_id': self.application_id,
            'guild_id': self.guild_id,
        }
        
        permission_overwrites = self.permission_overwrites
        if permission_overwrites is None:
            permission_overwrite_datas = []
        else:
            permission_overwrite_datas = [
                permission_overwrite.to_data() for permission_overwrite in permission_overwrites
            ]
        
        data['permissions'] = permission_overwrite_datas
        
        return data
    
    def __repr__(self):
        """Returns the application command permission's representation."""
        repr_parts = ['<', self.__class__.__name__, ' application_command_id=', repr(self.application_command_id),
            ' guild_id=', repr(self.guild_id), ', permission overwrite count=']
        
        permission_overwrites = self.overwrites
        if permission_overwrites is None:
            permission_overwrite_count = '0'
        else:
            permission_overwrite_count = repr(len(permission_overwrites))
        
        repr_parts.append(permission_overwrite_count)
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    def __eq__(self, other):
        """Returns whether the two application command permission's are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # No need to compare application_id, since `application_command_id` are already unique.
        if self.application_command_id != other.application_command_id:
            return False
        
        if self.guild_id != other.guild_id:
            return False
        
        if self.permission_overwrites != other.permission_overwrites:
            return False
        
        return True
    
    def __hash__(self):
        """Returns the application command overwrite's hash value."""
        hash_ = self.application_command_id ^ self.guild_id
        permission_overwrites = self.permission_overwrites
        if (permission_overwrites is not None):
            for permission_overwrite in permission_overwrites:
                hash_ ^= hash(permission_overwrite)
        
        return hash_
    
    def copy(self):
        """
        Copies the application command permission.
        
        Returns
        -------
        new : ``ApplicationCommandPermission``
        """
        new = object.__new__(type(self))
        
        new.application_id = self.application_id
        new.application_command_id = self.application_command_id
        new.guild_id = self.guild_id
        
        permission_overwrites = self.permission_overwrites
        if (permission_overwrites is not None):
            permission_overwrites = [permission_overwrite.copy() for permission_overwrite in permission_overwrites]
        
        new.permission_overwrites = permission_overwrites
        
        return new
    
    def add_permission_overwrite(self, permission_overwrite):
        """
        Adds an application command permission overwrite to the overwrites of the application command permission.
        
        Parameters
        ----------
        permission_overwrite : ``ApplicationCommandPermissionOverwrite``
            The overwrite to add.
        
        Raises
        ------
        AssertionError
            - If `overwrite` is not ``ApplicationCommandPermissionOverwrite`` instance.
            - If the application command permission has `10` overwrites already.
        """
        if __debug__:
            if not isinstance(permission_overwrite, ApplicationCommandPermissionOverwrite):
                raise AssertionError(f'`permission_overwrite` can be given as '
                    f'{ApplicationCommandPermissionOverwrite.__name__}  instance, got '
                    f'{permission_overwrite.__class__.__name__}.')
        
        permission_overwrites = self.permission_overwrites
        if permission_overwrites is None:
            self.permission_overwrites = permission_overwrites = []
        else:
            if __debug__:
                if len(permission_overwrites) >= APPLICATION_COMMAND_PERMISSION_OVERWRITE_MAX:
                    raise AssertionError(f'`permission_overwrites` can contain up to '
                        f'`{APPLICATION_COMMAND_PERMISSION_OVERWRITE_MAX}` permission_overwrites, which is already '
                        f'reached.')
        
        permission_overwrites.append(permission_overwrite)


class ApplicationCommandPermissionOverwrite:
    """
    Represents an application command's allow/disallow overwrite for the given entity.
    
    Attributes
    ----------
    allow : `bool`
        Whether the respective command is allowed for the represented entity.
    target_id : `int`
        The represented entity's identifier.
    target_type : ``ApplicationCommandPermissionOverwriteTargetType`
        The target entity's type.
    """
    __slots__ = ('allow', 'target_id', 'target_type')
    
    def __new__(cls, target, allow):
        """
        Creates a new ``ApplicationCommandPermission`` instance with the given parameters.
        
        Parameters
        ----------
        target : ``ClientUserBase`` or ``Role``, `tuple` ((``ClientUserBase``, ``Role`` type) or \
                `str` (`'Role'`, `'role'`, `'User'`, `'user'`), `int`)
            The target entity of the application command permission overwrite.
            
            The expected type & value might be pretty confusing, but the target was it to allow relaxing creation.
            To avoid confusing, here is a list of the expected structures:
            
            - ``Role`` instance
            - ``ClientUserBase`` instance
            - `tuple` (``Role`` type, `int`)
            - `tuple` (``ClientUserBase`` instance, `int`)
            - `tuple` (`'Role'`, `int`)
            - `tuple` (`'role'`, `int`)
            - `tuple` (`'User'`, `int`)
            - `tuple` (`'user'`, `int`)
        
        allow : `bool`
            Whether the respective application command should be enabled for the respective entity.
        
        Raises
        ------
        TypeError
            If `target` was not given as any of the expected types & values.
        AssertionError
            If `allow` was not given as `bool` instance.
        """
        # GOTO
        while True:
            if isinstance(target, Role):
                target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE
                target_id = target.id
                target_lookup_failed = False
                break
            
            if isinstance(target, ClientUserBase):
                target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER
                target_id = target.id
                target_lookup_failed = False
                break
            
            if isinstance(target, tuple) and len(target) == 2:
                target_maybe, target_id_maybe = target
                
                if isinstance(target_maybe, type):
                    if issubclass(target_maybe, Role):
                        target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE
                    elif issubclass(target_maybe, ClientUserBase):
                        target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER
                    else:
                        target_lookup_failed = True
                        break
                
                elif isinstance(target_maybe, str):
                    if target_maybe in ('Role', 'role'):
                        target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE
                    elif target_maybe in ('User', 'user'):
                        target_type = APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER
                    else:
                        target_lookup_failed = True
                        break
                
                else:
                    target_lookup_failed = True
                    break
                
                if type(target_id_maybe) is int:
                    target_id = target_id_maybe
                elif isinstance(target_id_maybe, int):
                    target_id = int(target_id_maybe)
                else:
                    target_lookup_failed = True
                    break
                
                target_lookup_failed = False
                break
            
            target_lookup_failed = True
            break
        
        if target_lookup_failed:
            raise TypeError(f'`target` can be given either as {Role.__name__}, {ClientUserBase.__name__}, '
                f'or as a `tuple` (({Role.__name__}, {User.__name__}, {UserBase.__name__} type or `str` '
                f'(`\'Role\'`, `\'role\'`, `\'User\'`, `\'user\'`)), `int`), got {target.__class__.__name__}: '
                f'{target!r}.')
        
        if __debug__:
            if not isinstance(allow, bool):
                raise AssertionError(f'`allow` can be given as `bool` instance, got {allow.__class__.__name__}.')
        
        self = object.__new__(cls)
        self.allow = allow
        self.target_id = target_id
        self.target_type = target_type
        return self
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new ``ApplicationCommandPermissionOverwrite`` instance from the received data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            The received application command permission overwrite data.
        
        Returns
        -------
        self : ``ApplicationCommandPermission``
            The created application command option.
        """
        self = object.__new__(cls)
        self.allow = data['permission']
        self.target_id = int(data['id'])
        self.target_type = ApplicationCommandPermissionOverwriteTargetType.get(data['type'])
        return self
    
    
    def to_data(self):
        """
        Converts the application command permission overwrite to a json serializable object.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        return {
            'permission': self.allow,
            'id': self.target_id,
            'type': self.target_type.value,
        }
    
    
    @property
    def target(self):
        """
        Returns the application command overwrite's target entity.
        
        Returns
        -------
        target : `None`, ``Role``, ``ClientUserBase``
        """
        target_type = self.target_type
        target_id = self.target_id
        
        if target_type is APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE:
            target = create_partial_role_from_id(target_id)
        
        elif target_type is APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER:
            target = create_partial_user_from_id(target_id)
        
        else:
            target = None
        
        return target
    
    
    def __repr__(self):
        """Returns the application command permission overwrite's representation."""
        return (
            f'<{self.__class__.__name__} target_type={self.target_type.name}, target_id={self.target_id!r}, '
            f'allow={self.allow!r}>'
        )
    
    
    def __eq__(self, other):
        """Returns whether the two application command overwrites are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.allow != other.allow:
            return False
        
        if self.target_type is not other.target_type:
            return False
        
        if self.target_id != other.target_id:
            return False
        
        return True
    
    
    def __hash__(self):
        """Returns the application command permission overwrite's hash value."""
        return self.target_type.value^(self.allow<<8)^self.target_id
    
    
    def copy(self):
        """
        Copies the application command permission overwrite.
        
        Returns
        -------
        new : ``ApplicationCommandPermissionOverwrite``
        """
        new = object.__new__(type(self))
        
        new.allow = self.allow
        new.target_type = self.target_type
        new.target_id = self.target_id
        
        return new
    
    
    def __gt__(self, other):
        """Returns whether self is greater than other."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_type_value = self.target_type.value
        other_type_value = other.target_type.value
        
        if self_type_value > other_type_value:
            return True
        
        if self_type_value < other_type_value:
            return False
        
        if self_type_value == APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE:
            self_target_id = self.target_id
            other_target_id = other.target_id
            
            self_role = ROLES.get(self_target_id, None)
            other_role = ROLES.get(other_target_id, None)
            if self_role is None:
                if other_role is None:
                    return (self_target_id > other_target_id)
                else:
                    return False
            else:
                if other_role is None:
                    return True
                else:
                    return (self_role > other_role)
        
        if self_type_value == APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER:
            return (self.target_id > other.target_id)
        
        # Should not happen
        return False
    
    
    def __lt__(self, other):
        """Returns whether self is greater than other."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_type_value = self.target_type.value
        other_type_value = other.target_type.value
        
        if self_type_value > other_type_value:
            return False
        
        if self_type_value < other_type_value:
            return True
        
        if self_type_value == APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_ROLE:
            self_target_id = self.target_id
            other_target_id = other.target_id
            
            self_role = ROLES.get(self_target_id, None)
            other_role = ROLES.get(other_target_id, None)
            if self_role is None:
                if other_role is None:
                    return (self_target_id < other_target_id)
                else:
                    return True
            else:
                if other_role is None:
                    return False
                else:
                    return (self_role < other_role)
        
        if self_type_value == APPLICATION_COMMAND_PERMISSION_OVERWRITE_TARGET_TYPE_USER:
            return (self.target_id < other.target_id)
        
        # Should not happen
        return True



__all__ = ('DiscordEntity', 'Slotted', )

from ...backend.export import include

id_to_datetime = include('id_to_datetime')


def _get_direct_parents_and_merge_slots(class_name, class_parents, class_attributes):
    """
    Gets the direct parent of the specified class definition and it's merged slots.
    
    Parameters
    ----------
    class_name : `str`
        The created class's name.
    class_parents : `tuple` of `type` instances
        The superclasses of the creates type.
    class_attributes : `dict` of (`str`, `Any`) items
        The class attributes of the created type.
    
    Returns
    -------
    direct_parent : `None` or `type`
        The direct parent of the respective class.
    final_slots : `set` of `str`
        Slots for the created type.
    """
    final_slots = set()
    
    parent_count = len(class_parents)
    if parent_count > 0:
        direct_parent = class_parents[0]
    else:
        direct_parent = None
    
    for class_parent in class_parents[1:]:
        if isinstance(class_parent, DiscordEntity):
            raise RuntimeError(f'`{class_name}` wanted to inherit `{DiscordEntity.__name__}` not as it\'s direct '
                f'(1st) parent type.')
    
    if (direct_parent is not None):
        # Subclasses might miss hash!
        if class_attributes.get('__hash__', None) is None:
            class_attributes['__hash__'] = direct_parent.__hash__
        
        # Remove weakref to avoid error
        try:
            final_slots.remove('__weakref__')
        except KeyError:
            pass
        
        index = 1
        while index < parent_count:
            class_parent = class_parents[index]
            final_slots.update(getattr(class_parent, f'_{class_parent.__name__}__slots', ()))
            index += 1
    
    slots = class_attributes.get('__slots__',)
    if (slots is not None) and slots:
        final_slots.update(slots)
    
    return direct_parent, final_slots


def _process_set_slot(class_attributes, final_slots):
    """
    Processes class attributes, which implement  `__set_slot__`.
    
    Parameters
    ----------
    class_attributes : `dict` of (`str`, `Any`) items
        The class attributes of the created type.
    final_slots : `set` of `str`
        Slots for the created type.
    """
    slotters = []
    for attribute_item in class_attributes.items():
        attribute_value = attribute_item[1]
        
        # Ignore types.
        if isinstance(attribute_value, type):
            continue
        
        # Check the type, whether it has the correct method.
        set_slot = getattr(type(attribute_value), '__set_slot__', None)
        if set_slot is None:
            continue
        
        # Queue up for future applications.
        slotters.append(attribute_item)
        continue
    
    # Apply slotters
    while slotters:
        attribute_name, slotter = slotters.pop()
        type(slotter).__set_slot__(slotter, attribute_name, class_attributes, final_slots)


class Slotted(type):
    """
    Metaclass for special slotted objects, which require `__set_slot__`.
    """
    def __new__(cls, class_name, class_parents, class_attributes):
        """
        Creates a new ``Slotted`` instance.
        
        Parameters
        ----------
        class_name : `str`
            The created class's name.
        class_parents : `tuple` of `type` instances
            The superclasses of the creates type.
        class_attributes : `dict` of (`str`, `Any`) items
            The class attributes of the created type.
        """
        direct_parent, final_slots = _get_direct_parents_and_merge_slots(class_name, class_parents, class_attributes)
        _process_set_slot(class_attributes, final_slots)
        class_attributes['__slots__'] = tuple(sorted(final_slots))
        
        return type.__new__(cls, class_name, class_parents, class_attributes)


class DiscordEntityMeta(Slotted):
    """
    Metaclass of Discord entities. Use ``DiscordEntity`` as superclass instead of using this class directly as a
    metaclass.
    """
    def __new__(cls, class_name, class_parents, class_attributes, immortal=False):
        """
        Creates a Discord entity type. Subclass ``DiscordEntity`` instead of using this class directly as a metaclass.
        
        Parameters
        ----------
        class_name : `str`
            The created class's name.
        class_parents : `tuple` of `type` instances
            The superclasses of the creates type.
        class_attributes : `dict` of (`str`, `Any`) items
            The class attributes of the created type.
        immortal : `bool`, Optional
            Whether the created type's instances should support weakreferencing. If the inherited type supports
            weakreferencing, then the subclass will as well of course. Defaults to `False`.
        
        Returns
        -------
        type : ``DiscordEntityMeta`` instance
        
        Notes
        -----
        The created instances are always slotted.
        
        When more classes are inherited then use `__slots` at the secondary classes for adding additional member
        descriptors.
        """
        direct_parent, final_slots = _get_direct_parents_and_merge_slots(class_name, class_parents, class_attributes)
        _process_set_slot(class_attributes, final_slots)
        
        if immortal:
            for class_parent in class_parents:
                if hasattr(class_parent, '__weakref__'):
                    break
            else:
                final_slots.add('__weakref__')
        
        if (direct_parent is not None) and ('id' not in class_attributes) and (direct_parent.id is DiscordEntity.id):
            final_slots.add('id')
        
        class_attributes['__slots__'] = tuple(sorted(final_slots))
        
        return type.__new__(cls, class_name, class_parents, class_attributes)


class DiscordEntity(object, metaclass=DiscordEntityMeta):
    """
    Base class for Discord entities.
    
    Notes
    -----
    Inherit it with passing `immortal = True` to make the subclass weakreferable.
    """
    @property
    def id(self):
        """
        Returns the discord entity's unique identifier number
        
        Returns
        -------
        id : `int`
        """
        return 0
    
    __slots__ = ()
    
    @property
    def created_at(self):
        """
        When the entity was created.
        
        Returns
        -------
        created_at : `datetime`
        """
        return id_to_datetime(self.id)
    
    def __hash__(self):
        """Returns the has value of the entity, what equals to it's id."""
        return self.id
    
    def __gt__(self, other):
        """Whether this entity's id is greater than the other's."""
        if type(self) is type(other):
            return self.id > other.id
        
        return NotImplemented
    
    def __ge__(self, other):
        """Whether this entity's id is greater or equal than the other's."""
        if type(self) is type(other):
            return self.id >= other.id
        
        return NotImplemented
    
    def __eq__(self, other):
        """Whether this entity's id is equal as the other's."""
        if type(self) is type(other):
            return self.id == other.id
        
        return NotImplemented
    
    def __ne__(self, other):
        """Whether this entity's id is not equal as the other's."""
        if type(self) is type(other):
            return self.id != other.id
        
        return NotImplemented
    
    def __le__(self, other):
        """Whether this entity's id is less or equal than the other's."""
        if type(self) is type(other):
            return self.id <= other.id
        
        return NotImplemented
    
    def __lt__(self, other):
        """Whether this entity's id is less than the other's."""
        if type(self) is type(other):
            return self.id < other.id
        
        return NotImplemented



__all__ = ()

from collections import deque

from ...env import CACHE_PRESENCE, CACHE_USER

from ...backend.utils import WeakReferer, set_docs
from ...backend.futures import Future, sleep, Task, WaitTillAll, CancelledError, future_or_timeout

from ..core import KOKORO
from ..gateway.client_gateway import REQUEST_MEMBERS as GATEWAY_OPERATION_CODE_REQUEST_MEMBERS
from ..events.intent import INTENT_MASK_GUILD_USERS, INTENT_MASK_GUILD_PRESENCES

READY_STATE_TO_DO_GUILD_IDS = set()
GUILD_RECEIVE_TIMEOUT = 5.0
SHARD_CONNECT_TIMEOUT = 12.0

USER_REQUEST_STATE_NONE = 0
USER_REQUEST_STATE_TIMEOUT = 1
USER_REQUEST_STATE_DONE = 2
USER_REQUEST_STATE_CANCELLED = 3
USER_REQUEST_STATE_ERROR = 4

class ShardUserRequester:
    """
    User requested task of a shard.
    
    Attributes
    ----------
    can_request_users : `bool`
        Whether the client can request users with it's gateway.
    gateway : ``DiscordGateway``
        The shard's gateway.
    guild_create_waiter : `None` or ``Future``
        Water to wait for guild create event. Used when no more guild id is received to use up.
    guild_ids : `set` of `int`
        The guild's id to request the users of.
    received_guild_ids : `deque` of `tuple` (`int`, `bool`)
        A queue of received tuples. Each tuple contains a guild's identifier and whether it's users should be
        requested.
    state : `int`
        A state containing the requester's state.
    task : `None` or ``Task`` of ``._runner``
        A task executing the user requesting.
    """
    __slots__ = ('can_request_users', 'gateway', 'guild_create_waiter', 'guild_ids', 'received_guild_ids', 'state',
        'task')
    
    def __new__(cls, gateway, guild_ids, can_request_users):
        """
        Creates a new user requester.
        
        Parameters
        ----------
        gateway : ``DiscordGateway``
            The shard's gateway.
        guild_ids : `set` of `int`
            The guild's id to request the users of.
        can_request_users : `bool`
            Whether the client can request users with it's gateway.
        """
        self = object.__new__(cls)
        self.can_request_users = can_request_users
        self.gateway = gateway
        self.guild_ids = guild_ids
        self.received_guild_ids = deque()
        self.guild_create_waiter = None
        self.state = USER_REQUEST_STATE_NONE
        self.task = Task(self._runner(), KOKORO)
        return self
    
    async def _runner(self):
        """
        Requests the users of the represented shard's guilds.
        
        This method is a coroutine.
        """
        try:
            guild_ids = self.guild_ids
            received_guild_ids = self.received_guild_ids
            can_request_users = self.can_request_users
            
            if can_request_users:
                sub_data = {
                    'guild_id': 0,
                    'query': '',
                    'limit': 0,
                    'presences': CACHE_PRESENCE,
                }
                
                data = {
                    'op': GATEWAY_OPERATION_CODE_REQUEST_MEMBERS,
                    'd': sub_data
                }
            
            while guild_ids:
                
                if not received_guild_ids:
                    guild_create_waiter = Future(KOKORO)
                    future_or_timeout(guild_create_waiter, GUILD_RECEIVE_TIMEOUT)
                    self.guild_create_waiter = guild_create_waiter
                    
                    try:
                        await guild_create_waiter
                    except TimeoutError:
                        self.state = USER_REQUEST_STATE_TIMEOUT
                        return
                    finally:
                        self.guild_create_waiter = None
                
                guild_id, should_request_users = received_guild_ids.popleft()
                guild_ids.discard(guild_id)
                
                if not can_request_users:
                    continue
                
                try:
                    READY_STATE_TO_DO_GUILD_IDS.remove(guild_id)
                except KeyError:
                    continue
                
                if not should_request_users:
                    continue
                
                sub_data['guild_id'] = guild_id
                await self.gateway.send_as_json(data)
                await sleep(0.6, KOKORO)
        
        except (CancelledError, GeneratorExit):
            self.state = USER_REQUEST_STATE_CANCELLED
            raise
        
        except:
            self.state = USER_REQUEST_STATE_ERROR
            raise
        
        else:
            self.state = USER_REQUEST_STATE_DONE
        
        finally:
            self.task = None
    
    
    def cancel(self):
        """
        Cancels the user request task.
        """
        task = self.task
        if (task is not None):
            self.task = None
            task.cancel()
    
    
    def feed(self, guild_id, should_request_users):
        """
        Feeds a guild identifier to the shard user requesters.
        
        Parameters
        ----------
        guild_id : `int`
            The guild's identifier.
        should_request_users : `bool`
            Whether the guild's users should be requested.
        
        Returns
        -------
        request_enqueued : `bool`
            Whether the request is queued up.
        """
        if guild_id in self.guild_ids:
            self.received_guild_ids.append((guild_id, should_request_users))
            
            request_enqueued = True
        else:
            request_enqueued = False
        
        guild_create_waiter = self.guild_create_waiter
        if (guild_create_waiter is not None):
            guild_create_waiter.set_result_if_pending(None)
        
        return request_enqueued
    
    def discard(self, guild_id):
        """
        Discards a guild's identifier from the shard user requester.
        
        Parameters
        ----------
        guild_id : `int`
            The guild's identifier.
        """
        self.guild_ids.discard(guild_id)


if CACHE_PRESENCE:
    def should_request_users_of(client, guild):
        if client.intents&INTENT_MASK_GUILD_PRESENCES:
            return guild.is_large
        else:
            return True
elif CACHE_USER:
    def should_request_users_of(client, guild):
        return True
else:
    def should_request_users_of(client, guild):
        return False

set_docs(should_request_users_of,
    """
    Returns whether the users of the given guild should be requested.
    
    Parameters
    ----------
    guild : ``Guild``
        A received guild instance.
    
    Returns
    -------
    client : ``Client``
        The respective client instance.
    should_request_users : `bool`
        Whether the guild's users should be requested.
    """)


class ReadyState:
    """
    Attributes
    ----------
    client_reference : ``WeakReferer`` to ``Client``
        Reference to the ready state's client.
    shard_count : `int`
        The amount of shards of the client.
    shard_user_requesters : `dict` of (`int`, ``ShardUserRequester``) items.
        User requester for each shard.
    shard_ready_waiter : `None` of ``Future``
        Waiter for shard ready event.
    task : `None` or ``Task`` to ``._runner``
        Task, which waits for all the user requests to finish.
    """
    __slots__ = ('client_reference', 'shard_count', 'shard_user_requesters', 'shard_ready_waiter', 'task')
    
    def __new__(cls, client):
        """
        Creates a new ready state instance from the given parameters.
        
        Parameters
        ----------
        client : ``Client``
            The respective client instance.
        """
        self = object.__new__(cls)
        self.shard_count = client.shard_count
        self.shard_user_requesters = {}
        self.task = Task(self._runner(), KOKORO)
        self.shard_ready_waiter = None
        self.client_reference = WeakReferer(client)
        
        return self
    
    def shard_ready(self, client, guild_datas, shard_id):
        """
        Sets the ready state's `.last_ready` to the current time and increases it's `.guild_left_counter` by the
        length of the given data.
        
        Parameters
        ----------
        guild_datas : `list` of `Any`
            Received guild datas.
        shard_id : `int`
            The shard's identifier.
        """
        shard_user_requesters = self.shard_user_requesters
        try:
            shard_user_requester = shard_user_requesters[shard_id]
        except KeyError:
            pass
        else:
            shard_user_requester.cancel()
        
        shard_count = client.shard_count
        gateway = client.gateway
        if shard_count:
            gateway = gateway.gateways[shard_id]
        
        guild_ids = set(int(guild_data['id']) for guild_data in guild_datas)
        
        can_request_users = client.intents&INTENT_MASK_GUILD_USERS
        
        if can_request_users:
            READY_STATE_TO_DO_GUILD_IDS.update(guild_ids)
        
        shard_user_requesters[shard_id] = ShardUserRequester(gateway, guild_ids, can_request_users)
        
        shard_ready_waiter = self.shard_ready_waiter
        if (shard_ready_waiter is not None):
            shard_ready_waiter.set_result_if_pending(None)
    
    def call_ready(self):
        """
        Calls the ready event handler of the respective client.
        """
        client = self.client_reference()
        if (client is not None):
            client.ready_state = None
            Task(client.events.ready(client), KOKORO)
    
    
    def feed_guild(self, client, guild):
        """
        Feeds the given `guild` to the ready state. Sets the last received guild's time to the current time and ends
        the ready state if there are no more guilds to receive.
        
        Parameters
        ----------
        client : ``Client``
            The respective client instance.
        guild : ``Guild``
            Received guild.
        
        Returns
        -------
        request_enqueued : `bool`
            Whether the request is queued up.
        """
        should_request_users = should_request_users_of(client, guild)
        
        guild_id = guild.id
        
        shard_count = self.shard_count
        if shard_count == 0:
            shard_id = 0
        else:
            shard_id = (guild_id>>22)%shard_count
        
        try:
            shard_user_requester = self.shard_user_requesters[shard_id]
        except KeyError:
            request_enqueued = False
        else:
            if shard_user_requester.state == USER_REQUEST_STATE_NONE:
                request_enqueued = shard_user_requester.feed(guild_id, should_request_users)
            else:
                request_enqueued = False
        
        return request_enqueued
    
    
    async def _runner(self):
        """
        Runner task of the ready state waiting for all guild users to be requested.
        
        This method is a coroutine.
        """
        try:
            shard_count = self.shard_count
            if not shard_count:
                shard_count = 1
            
            while True:
                tasks = None
                done_tasks = 0
                
                for shard_user_requester in self.shard_user_requesters.values():
                    if shard_user_requester.state == USER_REQUEST_STATE_NONE:
                        if tasks is None:
                            tasks = []
                        
                        tasks.append(shard_user_requester.task)
                    else:
                        done_tasks += 1
                
                if (tasks is not None):
                    await WaitTillAll(tasks, KOKORO)
                    continue
                
                if done_tasks == shard_count:
                    break
                
                shard_ready_waiter = Future(KOKORO)
                future_or_timeout(shard_ready_waiter, SHARD_CONNECT_TIMEOUT)
                self.shard_ready_waiter = None
                
                try:
                    await shard_ready_waiter
                except TimeoutError:
                    return
                finally:
                    self.shard_ready_waiter = None
        
        finally:
            self.task = None
            self.call_ready()
    
    def cancel(self):
        """
        Cancels the ready state.
        """
        task = self.task
        if (task is not None):
            self.task = None
            task.cancel()
            
            for shard_user_requester in self.shard_user_requesters.values():
                shard_user_requester.cancel()
    
    def __iter__(self):
        """
        Waits till the ready state receives all of it's shards and guilds, or till timeout occurs.
        
        This method is an awaitable generator.
        """
        task = self.task
        if (task is not None):
            try:
                yield from task
            except CancelledError:
                if self.task is not None:
                    raise
    
    __await__ = __iter__
    
    def discard_guild(self, guild):
        """
        Discards a guild from the guild requests.
        
        Parameters
        ----------
        guild : ``Guild``
            The guild to discard from the ready state.
        """
        guild_id = guild.id
        
        shard_count = self.shard_count
        if shard_count == 0:
            shard_id = 0
        else:
            shard_id = (guild_id>>22)%shard_count
        
        try:
            shard_user_requester = self.shard_user_requesters[shard_id]
        except KeyError:
            pass
        else:
            try:
                shard_user_requester.remove(guild_id)
            except KeyError:
                pass
            else:
                if guild.partial:
                    READY_STATE_TO_DO_GUILD_IDS.discard(guild_id)



__all__ = ('Achievement', )

from ..bases import DiscordEntity, IconSlot
from ..utils import DATETIME_FORMAT_CODE

from ..http import urls as module_urls

class Achievement(DiscordEntity):
    """
    Represents a Discord achievement created at Developer portal.
    
    Attributes
    ----------
    id : `int`
        The achievement's unique identifier number.
    application_id : `int`
        The achievement's respective application's id.
    description : `str`
        The description of the achievement.
    name : `str`
        The name of the achievement.
    secret : `bool`
        Secret achievements will *not* be shown to the user until they've unlocked them.
    secure : `bool`
        Secure achievements can only be set via HTTP calls from your server, not by a game client using the SDK.
    icon_hash : `int`
        The achievement's icon's hash. Achievements always have icon.
    icon_type : ``IconType``
        The achievement's icon's type.
    """
    __slots__ = ('application_id', 'description', 'name', 'secret', 'secure',)
    
    icon = IconSlot('icon', 'icon_hash', module_urls.achievement_icon_url, module_urls.achievement_icon_url_as)
    
    def __init__(self, data):
        """
        Creates an achievement with the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received achievement data.
        """
        self.application_id = int(data['application_id'])
        self.id = int(data['id'])
        
        self._update_attributes(data)
    
    def __repr__(self):
        """Returns the achievement's representation."""
        return f'<{self.__class__.__name__} id={self.id}, name={self.name!r}>'
    
    
    def __format__(self,code):
        if not code:
            return self.name
        
        if code=='c':
            return self.created_at.__format__(DATETIME_FORMAT_CODE)
        
        raise ValueError(f'Unknown format code {code!r} for object of type {self.__class__.__name__!r}')
    
    def _difference_update_attributes(self, data):
        """
        Updates the achievement and returns it's overwritten attributes as a `dict` with a `attribute-name` -
        `old-value` relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Achievement data received from Discord.
            
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        +---------------+-----------+
        | Keys          | Values    |
        +===============+===========+
        | name          | `str`     |
        +---------------+-----------+
        | description   | `str`     |
        +---------------+-----------+
        | secret        | `bool`    |
        +---------------+-----------+
        | secure        | `bool`    |
        +---------------+-----------+
        | icon          | ``Icon``  |
        +---------------+-----------+
        """
        old_attributes = {}
        
        name = data['name']['default']
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        description = data['description']['default']
        if self.description != description:
            old_attributes['description'] = self.description
            self.description = description
        
        secret = data['secret']
        if self.secret != secret:
            old_attributes['secret'] = self.secret
            self.secret = secret
        
        secure = data['secure']
        if self.secure != secure:
            old_attributes['secure'] = self.secure
            self.secure = secure
        
        self._update_icon(data, old_attributes)
        
        return old_attributes
    
    def _update_attributes(self, data):
        """
        Updates the achievement with overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Achievement data received from Discord.
        """
        self.name = data['name']['default']
        self.description = data['description']['default']
        
        self.secret = data['secret']
        self.secure = data['secure']
        
        self._set_icon(data)



__all__ = ('APPLICATIONS', 'APPLICATION_COMMANDS', 'BUILTIN_EMOJIS', 'CHANNELS', 'CLIENTS', 'DISCOVERY_CATEGORIES',
    'EMOJIS', 'EULAS', 'GUILDS', 'INTEGRATIONS', 'INVITES', 'KOKORO', 'MESSAGES', 'ROLES', 'SCHEDULED_EVENTS',
    'STAGES', 'STICKERS', 'STICKER_PACKS', 'TEAMS', 'UNICODE_TO_EMOJI', 'USERS')

from ..backend.utils import WeakValueDictionary, WeakKeyDictionary
from ..backend.event_loop import EventThread

__doc__ = """
Contains core cache and other objects relative to hata.

Caches
------
- `CLIENTS` : ``dict``
    
    Contains the created ``Client`` instances.
    
    Clients are only available for garbage collection after calling ``Client._delete``.


Weak Caches
-----------
- `APPLICATIONS` : ``WeakValueDictionary``
    
    Stores the requested or received ``Application`` instanced.
    
    Only here to avoid duplication.

- `APPLICATION_COMMANDS` : ``WeakValueDictionary``

    Stores the created ``ApplicationCommand`` instances.

- `CHANNELS` : ``WeakValueDictionary``
    
    Contains various ``ChannelBase`` instances.

- `DISCOVERY_CATEGORIES` : ``WeakValueDictionary``
    
    Stores the available ``DiscoveryCategory``.
    
    Discovery categories are populated startup time.

- `EMOJIS` : ``WeakValueDictionary``
    
    Contains all the emojis stored by the wrapper. Unicode emojis are never garbage collected, because they are present
    in their own specific containers, like `BUILTIN_EMOJIS` or `UNICODE_TO_EMOJI`.

- `EULAS` : ``WeakValueDictionary``
    
    Stores the ``Eula`` instances.
    
    When requesting an eula, querying from the cache is always preferred over requesting it.

- `GUILDS` : ``WeakValueDictionary``

    Contains the created ``Guild`` instances.
    
- `INTEGRATIONS` : ``WeakValueDictionary``
    
    Contains ``Integration`` instances.

- `INVITES` : ``WeakValueDictionary``
    
    Stores the created ``Invite`` instances. They keys are the invites' code.
    
    Only here to avoid duplication.

- `MESSAGES` : ``WeakValueDictionary``

    All non-deleted ``Message`` is stored inside of `MESSAGES`.
    
    Messages hve short life cycle, since their channel stores only a limited amount of instances (or non if configured
    like that). So make sure to store a reference to your message, if you want to keep it alive and receive events
    related to it.

- `ROLES` : ``WeakValueDictionary``
    
    Stores the created ``Role`` instanced.

- `SCHEDULED_EVENTS` : ``WeakValueDictionary``
    
    Storage for ``ScheduledEvent`` instances.
    
- `STAGES` : ``WeakValueDictionary``
    
    Stores the created ``Stage`` instanced.

- `STICKERS` : ``WeakValueDictionary``
    
    Stores the created ``Sticker`` instances.

- `STICKER_PACKS` : ``WeakValueDictionary``
    
    Storage for ``StickerPack``-s.

- `TEAMS` : ``WeakValueDictionary``
    
    Stores the created application ``Team`` instances.
    
    Only here to avoid duplication.

- `USERS` : ``WeakValueDictionary``
    
    Stores the created user instances.


Resolution Tables
-----------------
- `BUILTIN_EMOJIS` : `dict`
    
    Contains each builtin (unicode) ``Emoji`` by name.
    
    Version selector 16 emojis can be also access by putting `_vs16` postfix after their name.

- `UNICODE_TO_EMOJI` : `dict`
    
    Contains each builtin (unicode) ``Emoji`` by their unicode value.

Internal Resolution Tables
--------------------------
- `APPLICATION_ID_TO_CLIENT` : `dict`
    
    ``Application`` to it's respective ``Client`` relation.

- `INTERACTION_EVENT_RESPONSE_WAITERS` : ``WeakValueDictionary``
    
    Interactions waiting for their response message.
    
    Keys are the identifier of ``InteractionEvent``-s, meanwhile the values are the ``InteractionEvent``-s themselves.
    Used to resolve interaction's first response message, since that is not returned by the API directly.
    

- `INTERACTION_EVENT_MESSAGE_WAITERS` : ``WeakKeyDictionary``
    
    Interaction response message waiters.
    
    When a `INTERACTION_EVENT_RESPONSE_WAITERS` resolution succeeded, the interaction's response waiter future is
    accessed, which are stored in `INTERACTION_EVENT_MESSAGE_WAITERS` within ``InteractionEvent`` - ``Future``
    relation.

Immortal Objects
----------------
- KOKORO : ``EventThread``
    
    The asynchronous event loop & thread running hata clients.
"""

CLIENTS = {}

APPLICATIONS = WeakValueDictionary()
APPLICATION_COMMANDS = WeakValueDictionary()
CHANNELS = WeakValueDictionary()
DISCOVERY_CATEGORIES = WeakValueDictionary()
EMOJIS = WeakValueDictionary()
EULAS = WeakValueDictionary()
GUILDS = WeakValueDictionary()
INTEGRATIONS = WeakValueDictionary()
INVITES = WeakValueDictionary()
MESSAGES = WeakValueDictionary()
ROLES = WeakValueDictionary()
SCHEDULED_EVENTS = WeakValueDictionary()
STAGES = WeakValueDictionary()
STICKERS = WeakValueDictionary()
STICKER_PACKS = WeakValueDictionary()
TEAMS = WeakValueDictionary()
USERS = WeakValueDictionary()

APPLICATION_ID_TO_CLIENT = {}
INTERACTION_EVENT_RESPONSE_WAITERS = WeakValueDictionary()
INTERACTION_EVENT_MESSAGE_WAITERS = WeakKeyDictionary()

BUILTIN_EMOJIS = {}
UNICODE_TO_EMOJI = {}

KOKORO = EventThread(daemon=False, name='KOKORO')



__all__ = ('RateLimitProxy', )

from threading import current_thread

from ...backend.export import include
from ...backend.utils import WeakReferer
from ...backend.futures import Future
from ...backend.event_loop import LOOP_TIME

from .rate_limit import RateLimitGroup, LIMITER_GLOBAL, LIMITER_UNLIMITED, LIMITER_CHANNEL, LIMITER_WEBHOOK, \
    LIMITER_INTERACTION, LIMITER_GUILD, RateLimitHandler, UNLIMITED_SIZE_VALUE

ChannelBase = include('ChannelBase')
ChannelGuildBase = include('ChannelGuildBase')
Message = include('Message')
Role = include('Role')
Webhook = include('Webhook')
WebhookRepr = include('WebhookRepr')
Guild = include('Guild')
InteractionEvent = include('InteractionEvent')


class RateLimitProxy:
    """
    A proxy towards a rate limit.
    
    Attributes
    ----------
    _handler : None` or (`WeakReferer` to ``RateLimitHandler``)
        Reference to the actual rate limit handler of the `client` with the specified `group` and `limiter`.
    _key : ``RateLimitHandler``
        RateLimit handler used to lookup the active one.
    client : ``Client``
        Who's rate limits will be looked up.
    group : ``RateLimitGroup``
        The proxy's rate limit group to pull additional information.
    """
    __slots__ = ('_handler', '_key', 'client', 'group',)
    def __new__(cls, client, group, limiter=None, keep_alive=False):
        """
        Creates a new rate limit proxy.
        
        Parameters
        ----------
        client : ``Client``
            Who's rate limits will be looked up.
        group : ``RateLimitGroup``
            The proxy's rate limit group to pull additional information.
        limiter : ``DiscordEntity`` instance, Optional
            What's rate limits will be looked up.
            
            The accepted types depend on the group's limiter:
            +-----------------------+-----------------------------------------------------------------------+
            | Respective limiter    |   Accepted values                                                     |
            +=======================+=======================================================================+
            | LIMITER_CHANNEL       | ``ChannelBase``, ``Message``                                          |
            +-----------------------+-----------------------------------------------------------------------+
            | LIMITER_GUILD         | ``Guild``, ``ChannelGuildBase``, ``Message``, ``Role``, ``Webhook``,  |
            |                       | ``WebhookRepr``                                                       |
            +-----------------------+-----------------------------------------------------------------------+
            | LIMITER_WEBHOOK       | ``Webhook``, ``WebhookRepr``                                          |
            +-----------------------+-----------------------------------------------------------------------+
            | LIMITER_INTERACTION   | ``InteractionEvent``                                                  |
            +-----------------------+-----------------------------------------------------------------------+
            | LIMITER_GLOBAL        | `Any`                                                                 |
            +-----------------------+-----------------------------------------------------------------------+
            | LIMITER_UNLIMITED     | `Any`                                                                 |
            +-----------------------+-----------------------------------------------------------------------+
            
            Note that at the case of `LIMITER_GUILD` partial objects will yield `.guild` as `None` so `ValueError`
            will still be raised.
        keep_alive : `bool`, Optional
            Whether the rate limit proxy should keep alive the respective rate limit handler. Defaults to `False`.
        
        Raises
        ------
        RuntimeError
            If the given `group`'s limiter is not any of the predefined ones. Note that limiters are compared by memory
            address and not by value.
        TypeError
            If `group` was not given as ``RateLimitGroup`` instance.
        ValueError
            If the given `limiter` cannot be casted to `limiter_id` with the specified `group` .
        """
        if (type(group) is not RateLimitGroup):
            raise TypeError(f'`group` should be type `{RateLimitGroup.__name__}`, got {group}.__class__.__name__.')
        
        while True:
            group_limiter = group.limiter
            if group_limiter is LIMITER_GLOBAL:
                limiter_id = 0
                break
            
            elif group_limiter is LIMITER_UNLIMITED:
                limiter_id = 0
                break
            
            elif group_limiter is LIMITER_CHANNEL:
                if (limiter is not None):
                    if isinstance(limiter, ChannelBase):
                        limiter_id = limiter.id
                        break
                    
                    if isinstance(limiter, Message):
                        limiter_id = limiter.channel_id
                        break
            
            elif group_limiter is LIMITER_GUILD:
                if (limiter is not None):
                    if isinstance(limiter, Guild):
                        limiter_id = limiter.id
                        break
                    
                    if isinstance(limiter, (ChannelGuildBase, Message, Role, Webhook, WebhookRepr)):
                        
                        guild = limiter.guild
                        if (guild is not None):
                            limiter_id = limiter.id
                            break
            
            elif group_limiter is LIMITER_WEBHOOK:
                if (limiter is not None):
                    if isinstance(limiter, (Webhook, WebhookRepr)):
                        limiter_id = limiter.id
                        break
            
            elif group_limiter is LIMITER_INTERACTION:
                if (limiter is not None):
                    if isinstance(limiter, InteractionEvent):
                        limiter_id = limiter.id
                        break
            else:
                raise RuntimeError(f'`{group!r}.limiter` is not any of the defined limit groups.')
            
            raise ValueError(f'Cannot cast rate limit group\'s: `{group!r}` rate limit_id of: `{limiter!r}`.')
        
        key = RateLimitHandler(group, limiter_id)
        
        if keep_alive:
            key = client.http.handlers.set(key)
            handler = WeakReferer(key)
        else:
            handler = None
        
        self = object.__new__(cls)
        self.client = client
        self.group = group
        self._handler = handler
        self._key = key
        return self
    
    def is_limited_by_channel(self):
        """
        Returns whether the represented rate limit group is limited by channel id.
        
        Returns
        -------
        is_limited_by_channel : `bool`
        """
        return (self.group.limiter is LIMITER_CHANNEL)
    
    def is_limited_by_guild(self):
        """
        Returns whether the represented rate limit group is limited by guild id.
        
        Returns
        -------
        is_limited_by_guild : `bool`
        """
        return (self.group.limiter is LIMITER_GUILD)
    
    def is_limited_by_webhook(self):
        """
        Returns whether the represented rate limit group is limited by webhook id.
        
        Returns
        -------
        is_limited_by_webhook : `bool`
        """
        return (self.group.limiter is LIMITER_WEBHOOK)
    
    def is_limited_by_interaction(self):
        """
        Returns whether the represented rate limit group is limited by interaction id.
        
        Returns
        -------
        is_limited_by_interaction : `bool`
        """
        return (self.group.limiter is LIMITER_INTERACTION)
    
    def is_limited_globally(self):
        """
        Returns whether the represented rate limit group is limited globally,
        
        Returns
        -------
        is_limited_globally : `bool`
        """
        return (self.group.limiter is LIMITER_GLOBAL)
    
    def is_unlimited(self):
        """
        Returns whether the represented rate limit group is unlimited,
        
        Returns
        -------
        is_unlimited : `bool`
        """
        return (self.group.limiter is LIMITER_UNLIMITED)
    
    def is_alive(self):
        """
        Returns whether the respective client has the represented rate limit handler is alive.
        
        Returns
        -------
        is_alive : `bool`
        """
        return (self.handler is not None)
    
    def has_info(self):
        """
        Returns whether the represented rate limit handler now stores any rate limit information.
        
        Returns
        -------
        has_info : `bool`
        """
        handler = self.handler
        if handler is None:
            return False
        
        return (handler.queue is not None)
    
    @property
    def keep_alive(self):
        """
        Get-set property for accessing whether the rate limit proxy should keep alive the respective rate limit
        handler.
        
        Accepts and returns `bool`.
        """
        handler = self._handler
        if handler is None:
            return False
        
        return (handler() is self._key)
    
    @keep_alive.setter
    def keep_alive(self, value):
        if value:
            while True:
                handler = self._handler
                if handler is None:
                    break
                
                handler = handler()
                if handler is None:
                    break
                
                if (self._key is not handler):
                    self._key = handler
                
                return
            
            key = self._key
            handler = self.client.http.handlers.set(key)
            if (handler is not key):
                self._key = handler
            self._handler = WeakReferer(handler)
            return
        else:
            handler = self._handler
            if handler is None:
                return
            
            handler = handler()
            if handler is None:
                return
            
            if self._key is handler:
                self._key = handler.copy()
            return
    
    @property
    def limiter_id(self):
        """
        Returns the the represented rate limit handler's `.limiter_id`.
        
        Returns
        -------
        limiter_id : `int`
        """
        return self._key.limiter_id
    
    def has_size_set(self):
        """
        Returns whether the represented rate limit group size is already set.
        
        Not only not used rate limit groups, but still optimistic or unlimited rate limit groups fall under this
        category as well.
        
        Returns
        -------
        has_size_set : `bool`
        """
        return (self.group.size > 0)
    
    @property
    def size(self):
        """
        Returns the represented rate limit group's size.
        
        Returns
        -------
        size : `int`
        """
        return self.group.size
    
    @property
    def handler(self):
        """
        Returns the client's represented active handler if applicable.
        
        Returns
        -------
        handler : `None` or ``RateLimitHandler``
        """
        handler = self._handler
        if (handler is not None):
            handler = handler()
            if (handler is not None):
                return handler
        
        handler = self.client.http.handlers.get(self._key)
        if (handler is not None):
            self._handler = WeakReferer(handler)
        
        return handler
    
    @property
    def used_count(self):
        """
        Returns how much requests are used up or already done right now.
        
        Returns
        -------
        count : `int`
        """
        handler = self.handler
        if handler is None:
            return 0
        
        return (handler.active + handler.count_drops())
    
    @property
    def free_count(self):
        """
        Returns how much requests can be done towards the represented rate limit.
        
        If the rate limit proxy represents an unlimited endpoint, then `0` is returned.
        
        Returns
        -------
        free_count : `int`
        """
        size = self.group.size
        if size < 1:
            if size == 0:
                size = 1
            elif size == UNLIMITED_SIZE_VALUE:
                return 0
            else:
                size = -size
        
        handler = self.handler
        if handler is None:
            return size
        
        return (size - handler.active - handler.count_drops())
    
    @property
    def waiting_count(self):
        """
        Returns how much requests are waiting in queue by the represented rate limit.
        
        Returns
        -------
        count : `int`
        """
        handler = self.handler
        if handler is None:
            return 0
        
        queue = handler.queue
        if queue is None:
            return 0
        
        return len(queue)
    
    def __hash__(self):
        """Hashes the rate limit proxy."""
        return self.group.group_id^self._key.limiter_id
    
    class _wait_till_limits_expire_callback:
        """
        `WeakReferer` callback used at ``.wait_till_limits_expire`` for waking it up.
        """
        __slots__ = ('future')
        def __init__(self, future):
            self.future = future
        
        def __call__(self, reference):
            future = self.future
            future.set_result_if_pending(None)
            loop = future._loop
            if current_thread() is not loop:
                loop.wake_up()
    
    async def wait_till_limits_expire(self):
        """
        Waits till the represented rate limits expire.
        
        This method is a coroutine.
        
        Raises
        ------
        RuntimeError
            If the method is called meanwhile `keep_alive` is `True`.
        
        Notes
        -----
        The waiting is implemented with weakreference callback, so the coroutine returns when the source callback is
        garbage collected. This also means waiting on the exact same limit multiple times causes misbehaviour.
        """
        handler = self._handler
        
        while True:
            if (handler is not None):
                handler = handler()
                if (handler is not None):
                    break
            
            handler = self.client.http.handlers.get(self._key)
            if handler is None:
                return
            
            break
        
        if handler is self._key:
            raise RuntimeError('Cannot use `.wait_till_limits_expire` meanwhile `keep_alive` is `True`.')
        
        future = Future(current_thread())
        self._handler = WeakReferer(handler, self._wait_till_limits_expire_callback(future))
        await future
    
    @property
    def next_reset_at(self):
        """
        Returns when the next rate limit reset will happen of the represented rate limit handler.
        
        If there is no active rate limit handler represented or if the handler has has 0 used up limits, then returns
        `0.0`. if there is any, then returns it in `LOOP_TIME` time.
        
        Returns
        -------
        next_reset_at : `float`
        """
        handler = self.handler
        if handler is None:
            return 0.0
        
        drops = handler.drops
        if (drops is None) or (not drops):
            return 0.0
        
        return drops[0].drop
    
    @property
    def next_reset_after(self):
        """
        Familiar to ``.next_reset_at`` but it instead returns how time is left till next reset instead.
        
        Returns
        -------
        next_reset_after : `float`
        """
        handler = self.handler
        if handler is None:
            return 0.0
        
        drops = handler.drops
        if (drops is None) or (not drops):
            return 0.0
        
        return drops[0].drop-LOOP_TIME()



__all__ = ('MessageActivity', )

from .preinstanced import MessageActivityType

class MessageActivity:
    """
    Might be sent with a ``Message``, if it has rich presence-related chat embeds.
    
    Attributes
    ----------
    party_id : `str`
        The message application's party's id. Can be empty string.
    type : ``MessageActivityType``
        The message application's type.
    """
    __slots__ = ('party_id', 'type',)
    def __init__(self, data):
        """
        Creates a new ``MessageActivity`` from message activity data included inside of a ``Message``'s data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Message activity data.
        """
        self.party_id = data.get('party_id','')
        self.type = MessageActivityType.get(data['type'])

    def __eq__(self, other):
        """Returns whether the two message activities are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.type is not other.type:
            return False
        
        if self.party_id != other.party_id:
            return False
        
        return True
    
    def __repr__(self):
        """Returns the message activity's representation."""
        return f'<{self.__class__.__name__} type={self.type.name} ({self.type.value}), party_id={self.party_id!r}>'



__all__ = ('WelcomeChannel', 'WelcomeScreen')

import reprlib

from ...backend.utils import BaseMethodDescriptor

from ..core import CHANNELS
from ..channel import ChannelText, ChannelBase
from ..emoji import Emoji, create_partial_emoji_from_data
from ..preconverters import preconvert_snowflake


class WelcomeScreen:
    """
    Represents a guild's welcome screen.
    
    Attributes
    ----------
    description : `None` or `str`
        Description, of what is the server about.
    welcome_channels : `None` or `tuple` of ``WelcomeChannel``
        The featured channels by the welcome screen.
    """
    __slots__ = ('description', 'welcome_channels', )
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new welcome screen instance from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Welcome screen data.
        """
        description = data.get('description', None)
        if (description is not None) and (not description):
            description = None
        
        welcome_channel_datas = data.get('welcome_channels', None)
        if (welcome_channel_datas is None) or (not welcome_channel_datas):
            welcome_channels = None
        else:
            welcome_channels = tuple(
                WelcomeChannel.from_data(welcome_channel_data) for welcome_channel_data in welcome_channel_datas
            )
        
        self = object.__new__(cls)
        self.description = description
        self.welcome_channels = welcome_channels
        return self
    
    
    def to_data(self):
        """
        Converts the welcome screen to a json serializable object.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        return {
            'description': self.description,
            'welcome_channels': [welcome_channel.to_data() for welcome_channel in self.welcome_channels],
        }
    
    def __repr__(self):
        """Returns the welcome screen's representation."""
        return (f'<{self.__class__.__name__} description={reprlib.repr(self.description)}, welcome_channels='
            f'{self.welcome_channels!r}>')
    
    def __hash__(self):
        """Returns the welcome screen's hash."""
        return hash(self.description) ^ hash(self.welcome_channels)
    
    def __eq__(self, other):
        """Returns whether the two welcome screens are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.description != other.description:
            return False
        
        if self.welcome_channels != other.welcome_channels:
            return False
        
        return True


class WelcomeChannel:
    """
    Represents a featured channel by a welcome screen.
    
    Attributes
    ----------
    description : `str`
        The channel's short description.
    channel_id : `int`
        The channel's id.
    emoji : ``Emoji``
        The emoji displayed before the `description`.
    """
    __slots__ = ('description', 'channel_id', 'emoji', )
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new welcome channel instance from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Welcome channel data.
        """
        self = object.__new__(cls)
        self.channel_id = int(data['channel_id'])
        self.description = data['description']
        self.emoji = create_partial_emoji_from_data(data)
        return self
    
    
    def to_data(self):
        """
        Converts the welcome channel to a json serializable object.
        
        Returns
        -------
        data : `str`
        """
        data = {
            'channel_id': self.channel.id,
            'description': self.description,
        }
        
        emoji = self.emoji
        if emoji.is_unicode_emoji():
            emoji_id = None
            emoji_name = emoji.unicode
        else:
            emoji_id = emoji.id
            emoji_name = emoji.name
        
        data['emoji_id'] = emoji_id
        data['emoji_name'] = emoji_name
        
        return data
    
    
    @BaseMethodDescriptor
    def custom(cls, base, **kwargs):
        """
        Creates a custom welcome channel. If called as a classmethod, then all parameters are required tho if called
        from an instance, then only those should be given, which you intent to modify.
        
        Parameters
        ----------
        **kwargs : keyword parameters
            Additional attributes of the created welcome channel.
        
        Other Parameters
        ----------------
        channel : ``ChannelTextBase`` or `int` instance, Optional
            The channel of the welcome screen.
        channel_id : `int`, optional
            Alias of `channel`, tho it accepts only snowflake.
            
            Mutually exclusive with the `channel` parameter.
        description : `str`, Optional
            Description of the welcome screen.
        emoji : ``Emoji``, Optional
            The emoji of the welcome screen.
        
        Returns
        -------
        self : ``WelcomeChannel``
        
        Raises
        ------
        TypeError
            - If `channel` parameter was given as a channel, but not as ``ChannelText`` instance.
            - If `channel` parameter was not given neither as ``ChannelText`` or `int` instance.
            - If `channel_id` was given but neither as `int` or `str` instance.
            - If `description` was not given as `str` instance.
            - If `emoji` was not given as ``Emoji`` instance.
        ValueError
            - If `channel` was given as `str` instance, but not convertable to `int`.
            - If `channel` was given as `int` instance, but out of the expected range.
            - If `channel_id` was given as `str` instance, but not convertable to `int`.
            - If `channel_id` was given as `int` instance, but out of the expected range.
            - If `description` was given as empty string.
        """
        while True:
            try:
                channel = kwargs.pop('channel')
            except KeyError:
                pass
            else:
                if isinstance(channel, ChannelText):
                    channel_id = channel.id
                elif isinstance(channel, ChannelBase):
                    raise TypeError(f'`channel` parameters can be given as {ChannelText.__name__} or `int` instance,'
                        f'got an other channel type, {channel.__class__.__name__}.')
                else:
                    channel_id = preconvert_snowflake(channel, 'channel')
                
                break
            
            try:
                channel_id = kwargs.pop('channel_id')
            except KeyError:
                pass
            else:
                channel_id = preconvert_snowflake(channel_id, 'channel_id')
                break
            
            if base is None:
                raise TypeError(f'`channel` or `channel_id` are required parameters if `{cls.__name__}.custom` is '
                    f'called as a classmethod.')
            
            channel_id = base.channel_id
            break
        
        try:
            description = kwargs.pop('description')
        except KeyError:
            if base is None:
                raise TypeError(f'`description` is a required parameter if `{cls.__name__}.custom` is called as a '
                    f'classmethod.') from None
            
            description = base.description
        else:
            if not isinstance(description, str):
                raise TypeError(f'`description` can be given as `str` instance, got {description.__class__.__name__}.')
            
            if not description:
                raise ValueError(f'`description` cannot be given as empty string.')
            
        try:
            emoji = kwargs.pop('emoji')
        except KeyError:
            if base is None:
                raise TypeError(f'`emoji` is a required parameter if `{cls.__name__}.custom` is called as a '
                    f'classmethod.') from None
            
            emoji = base.emoji
        else:
            if not isinstance(emoji, Emoji):
                raise TypeError(f'`emoji` can be given as `{Emoji.__name__}` instance, got {emoji.__class__.__name__}.')
        
        
        if kwargs:
            raise TypeError(f'Unused parameters: {", ".join(list(kwargs))}')
        
        self = object.__new__(cls)
        self.channel_id = channel_id
        self.description = description
        self.emoji = emoji
        
        return self
    
    
    @property
    def channel(self):
        """
        Returns the welcome channel's respective channel.
        
        Returns
        -------
        channel : ``ChannelText``
        """
        channel_id = self.channel_id
        try:
            channel = CHANNELS[channel_id]
        except KeyError:
            channel = ChannelText._from_partial_data(None, channel_id, None)
        
        return channel
    
    
    def __repr__(self):
        """Returns the welcome channel's representation."""
        return (f'<{self.__class__.__name__} channel_id={self.channel_id},  emoji={self.emoji!r}, description='
            f'{reprlib.repr(self.description)}>')
    
    
    def __hash__(self):
        """Returns the welcome channel's hash."""
        return self.channel_id ^ self.emoji.id ^ hash(self.description)
    
    
    def __eq__(self, other):
        """Returns whether the two welcome channels are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.channel_id != other.channel_id:
            return False
        
        if (self.emoji is not other.emoji):
            return False
        
        if self.description != other.description:
            return False
        
        return True
